//
// Initialization
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <stdlib.h>
#include <guit.hpp>
#include <gfile.hpp>
#include <gtheme.hpp>
#include <gcolor.hpp>
#include <ghotkey.hpp>
#include <gvalue.hpp>
#include <gmodes.hpp>
#include <gcontrolimpl.hpp>
#include <gstyle.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

#if defined(__APPLE__)
#  define GUIT_DIR "Library/Guit/"

#elif defined(_WIN32)
#  define GUIT_DIR "Guit/"

#else   // Linux/Unix/FreeBSD,etc.
#  define GUIT_DIR ".guit/"
#endif


int GCore::argc_{};
const char** GCore::argv_{};
GTheme* GCore::theme_{};
class GInspector* GCore::inspector_{};
int GCore::inspecting_{};
Gadget* GCore::inspected_{};
std::function<void(Gadget*)> GCore::doinspect_{};


GFatalError::GFatalError(GString const& where, GString const& msg, GObject const* obj)
: what_(where + ": " + msg) {
  gerror(where, msg, obj);
}


GPrefs::GPrefs() {
#if defined(__APPLE__)
  fontdirs_ = {
    "~/Library/Fonts/",
    "/Library/Fonts/",
    "/System/Library/Fonts/",
    "/System/Library/Fonts/Supplemental/",
    //"/Network/Library/Fonts/"
  };

#elif defined(__linux__) || defined(__unix__)
  /*
  fontdirs_ = {       fontdirs_ is not set => fc-list will be called
    "~/.fonts/",
    "~/.local/share/fonts",
    "/usr/share/fonts/truetype/",
    "/usr/share/fonts",
    "/usr/X11R6/lib/X11/fonts/ttfonts/",
    "/usr/X11R6/lib/X11/fonts/",
    "/usr/local/share/fonts"
  };*/
#elif defined(_WIN32) || defined(_WIN64)
  fontdirs_ = {"C:/Windows/Fonts"};
#else

#error "Unknown OS"

#endif

  for (auto& it : fontdirs_) {
    it = normPath(it, true);
  }
  
  sansfonts_ = {
    "Arial", "Helvetica", "Roboto", "Verdana", //"DejaVu Sans",
    "Liberation Sans", "FreeSans", "Ubuntu",
  };

  seriffonts_= {
    "Times", "Times New Roman", "Georgia", //"DejaVu Serif",
    "Liberation Serif",
  };

  monofonts_ = {
    "Courrier", "Courrier New", "Courrier Prime", //"DejaVu Sans Mono",
    "Liberation Mono", "FreeMono", "Ubuntu Mono",
  };

  selectionColor_ = new GRgba(173.f/255.f, 216.f/255.f, 230.f/255.f, 1.f);
  caretColor_ = new GRgba(1., 0, 0);
  inspectcolor_ = &GColor::guitblue;
  inspectedcolor_ = &GColor::orange;
  inspecthotkey_ = GEventState::Meta;  // | Event::CtrlMode;
}


int GPrefs::precision() {return 3;}
float GPrefs::wheelGain() {return -0.025f;}     // TO PARAMETERIZE
GTime GPrefs::tipDelay() {return 1000;}
GTime GPrefs::dblclickDelay() {return 300;}
GTime GPrefs::animTimeout() {return 25;}
float GPrefs::caretWidth () {return 1.f;}
float GPrefs::caretEndDist () {return 5.f;}
int   GPrefs::caretScroll() {return 5;}
float GPrefs::iconWidth() {return 20.f;}
float GPrefs::menuTab() {return 20.f;}
float GPrefs::resizeHandleSize() {return 9.f;}
float GPrefs::resizeFrameShift() {return 6.f;}  // DEPENDANCE !!!


GRgba const& GPrefs::resizeFrameColor() {
  static GRgba c(17.f/255.f, 117.f/255.f, 217.f/255.f, 0.6f);
  return c;
}

GRgba const& GPrefs::resizeHandleColor() {
  static GRgba c(17.f/255.f, 117.f/255.f, 217.f/255.f, 0.3f);
  return c;
}

GString const& GPrefs::openDir() {
  static GString d = "~/";
  return d;
}

GString const& GPrefs::saveDir() {
  static GString d = "~/";
  return d;
}

GStrings const& GPrefs::fontDirs() {
  return GControlImpl::instance().fontdirs_;
}

GStrings const& GPrefs::sansFonts() {
  return GControlImpl::instance().sansfonts_;
}

GStrings const& GPrefs::serifFonts() {
  return GControlImpl::instance().seriffonts_;
}

GStrings const& GPrefs::monoFonts() {
  return GControlImpl::instance().monofonts_;
}

GRgba const& GPrefs::selectedTextColor() {
  return *GControlImpl::instance().selectionColor_;
}

GRgba const& GPrefs::caretColor() {
  return *GControlImpl::instance().caretColor_;
}

GColor &GPrefs::inspectColor() {
  return *GControlImpl::instance().inspectcolor_;
}

GColor &GPrefs::inspectedColor() {
  return *GControlImpl::instance().inspectedcolor_;
}

GTheme& GPrefs::theme() {
  if (!GCore::theme_) GCore::theme_ = new GDefaultTheme();
  return *GCore::theme_;
}

GCore::GCore() {

  // il faut initialiser les props sinon on les trouvera pas toutes dans les CSS
  GAlign::Type();
  GBorder::Type();
  GColor::Type();
  GBackground::Type();
  GFlex::Type();
  GFlow::Type();
  GFont::Type();
  GFormat::Type();
  GGap::Type();
  GMargin::Type();
  GPos::Type();
  GSize::Type();

  GShow::Type();
  GCheck::Type();
  GEnable::Type();
  GModes::Type();

  GBool::Type();
  GInt::Type();
  GFloat::Type();
  GText::Type();
  GTip::Type();
  GChoice::Type();
  GIcon::Type();
  GHotkey::Type();
  GValue::Type();
  GItems::Type();
  GTextData::Type();

  GAnim::Type();
}

// - - - App - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GApp::GApp(int argc, const char** argv) {
  GControlImpl::instance().obtainChannel(0);;
  GCore::argc_ = argc;
  GCore::argv_ = argv;
  //CAUTION: theme and styles **MUST** be created **AFTER** GNatApp
  GCore::theme_ = new GDefaultTheme();
  GCore::theme_->init();
}

GDim GApp::screenSize() {
  return GCore::natApp()->screenSize();
}

Gadget* GApp::currentFocus() {
  return GCore::obtainChannel(0).focusg_.get();
}

bool GApp::addCssFile(GString const& css_file) {
  GIFile in(css_file);
  if (in) return GStyleBase::instance().load(in, css_file);
  else {
    gerror("GApp::addCssFile", "Can't find CSS file: " + css_file);
    return false;
  }
}

bool GApp::addCssRules(GString const& css_rules) {
  return GStyleBase::instance().loadBuffer(css_rules);
}

GApp& GApp::operator<<(GString const& css) {
  if (css.find('{') != GString::npos) addCssRules(css);
  else addCssFile(css);
  return *this;
}

// - - - GApp Resources - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
GString const& GApp::homeDir() {
  static GString dir = normPath("~", true);
  return dir;
}
GString const& GApp::desktopDir() {
  static GString dir = normPath("~/" DESKTOP_DIR, true);
  return dir;
}*/

GString const& GApp::guitDir() {
  static GString dir = normPath("~/" GUIT_DIR, true);
  static bool init{};
  if (!init) {
    init = true;
    if (!fileExist(dir)) {
      if (!makeDir(dir)) gerror("GApp::guitDir","~/" GUIT_DIR " could not be created");
    }
  }
  return dir;
}

// Note: GCore::arg must have been init!
GString const& GApp::appDir() {
  static GString dir = normPath("~/" GUIT_DIR + execName(), true);
  static bool init{};
  if (!init) {
    init = true;
    if (!fileExist(dir)) {
      if (!makeDir(dir)) gerror("GApp::appDir",
                                "~/" GUIT_DIR + execName() + " could not be created");
    }
  }
  return dir;
}

GStrings const& GApp::resourceDirs() {
  static GStrings* dirs{};
  if (!dirs) {
    dirs = new GStrings();

#if defined(__APPLE__)
    if (GCore::argc_ > 0) {     // find Application own resources if any
      GString d = execPath();
      auto pos = d.find("MacOS/");
      if (pos != GString::npos) {
        d.replace(pos, GString::npos, "Resources/");
        dirs->push_back(d);
      }
    }
#endif
    dirs->push_back(appDir());

    std::clog << "Guit: launching " << execName() << "\n"
    << "Resource files:\n";
    for (auto& it :* dirs) cerr << "- " << it << "\n";
    std::clog << endl;
  }
  return *dirs;
}

GString GApp::execPath() {
  return GCore::argc_ > 0 ? normPath(GCore::argv_[0]) : "";
}

GString GApp::execName() {
  return basename(execPath(), false);
}

int GApp::argc() {
  return GCore::argc_;
}

GString GApp::argv(int n) {
  return (n >= 0 && n < GCore::argc_) ? GCore::argv_[n] : "";
}

bool GApp::hasArg(GString const& name) {
  if (name.empty()) return false;
  for (int k = 1; k < GCore::argc_; ++k) {
    if (name == GCore::argv_[k]) return true;
  }
  return false;
}

bool GApp::hasArg(GString const& name, GString& value, GString const& assign) {
  value.clear();
  if (name.empty()) return false;
  if (assign.empty()) return hasArg(name);

  for (int k = 1; k < GCore::argc_; ++k) {
    GString arg(GCore::argv_[k]);
    GString argname, argval;
    strsplit(arg, argname, argval, assign);
    if (argname == name) {value = argval; return true;}
  }
  return false;
}

}
