//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

@page gadgetdoc
/// Gadget is the base class of all graphical objects. Its main subclass is GBox, the base class
/// of gadgets that can have chid gadgets.
///
/// @section Adding objects
///
/// Other objects (gadgets or properties) can be added using the << operator (see operator<<())
/// as in the following example:
/// @code
///    auto& mymenu = Menu    //  A menu containing 3 items
///    << [&]{ doIt(on::item()) ;}
///    << "New..."
///    << "Open..."
///    << "Save...";
/// @endcode
/// This example creates a menu to which are added a callback function and 3 items:
/// - mymenu is a C++ reference (& symbol) referencing the menu. Its type is inferred by the
///  compiler (auto keyword).
/// - The callback is a C++ lambda that captures variables ([&] syntax). It will be called each time
///  a menu item is clicked. on::item() returns the item that was clicked.
/// - The 3 items are specified as strings that specify the title they will display. This syntax
///  auto-creates 3 Item gadgets. Such items are auto-adaptative, meaning they adapt their
///  look and feel to their parent. Thus, they will behave and look as menu items.
///
/// The << operator is polymorphic. It always means "add to" and performs different low-level
/// operations depending on its operands. Properties and callbacks can be added using this
/// operator, for instance:
/// @code
///    auto& mymenu = Menu
///    << ( "New..." << Color("red") << Font("bold 16px Arial") )
///    // << etc.
/// @endcode
/// adds a Color and a Font property to the "New..." item, so that it will be displayed in red using this font.
/// Note the parentheses around "New..", without them, the Color and the Font would be added to
/// the Menu.
///
/// An abritrary hierarchy of objects can created in this way, here a MenuBar that will opens mymenu
/// and a second menu:
/// @code
///   auto& mbar = MenuBar     // A MenuBar opening 2 Menus
///   << ("File" << mymenu)
///   << ("Edit"
///       << (Menu
///           << ("Cut"    << [&]{ doCut(); })
///           << ("Paste"  << [&]{ doPaste(); })
///           << Separator
///           << ("Delete" << [&]{ doDelete(); })
///          )
///      );
/// @endcode
/// "File" and "Edit" are the items of the MenuBar. Each of them opens the Menu that it added to it.
/// The "Cut", "Paste" and "Delete" items have their own callbacks (a callback could also be added
/// to the "Edit" Menu, as in the previous example).
///
/// @section Box and non-Box gadgets
///
/// Any gadget can be added to a Box gadget (such as Menu or MenuBar). Only Menus and Windows
/// can be added to non-Box gadgets (such as Items). Box gadgets are implemented using the GBox
/// class (and its subclasses). GBox derives from Gadget.
///
/// All boxes are created in the same way as above:
/// @code
///    auto& mylist = ListBox       // Item List
///    << "Left"
///    << "Center"
///    << "Right";
///
///   auto& myradios = RadioBox     // Radio buttons
///    << "Left"
///    << "Center"
///    << "Right";
///
///    auto& mychoices = ChoiceBox   // State buttons
///    << "Left"
///    << "Center"
///    << "Right";
///
///    auto& myoptions = OptionBox   // Opens a menu of options
///    << "Left"
///    << "Center"
///    << "Right";
///
///    auto& mycombo = ComboBox     // Combo box
///    << "Left"
///    << "Center"
///    << "Right";
/// @endcode
/// Contrary to a Menu or a MenuBar, these boxes enforce exclusive selection: clicking on an item
/// selects it. As above, the items auto-adapt to the look and feel of their parent.
///
/// Callbacks also adapt: the callbacks of these boxes are called when an item is selected
/// (interactively or programmatically), while the callbacks of the Menu and the MenuBar are called
/// when an item is clicked, as these boxes do not enforce exclusive selection.
///
/// All boxes have a similar API, meaning that can not only be created, but also managed almost
/// in the same way. ListBox, RadioBox, ChoiceBox have exactly the save API, OptionBox and
/// ComboBox have almost the same API (the only difference is that they provide access to their
/// menu and selector). Many boxes are thus totally or almost totally interchangeable.
///
/// @section VBox, HBox and Specific Items
///
/// When a specific look and feel is needed, one can specify the type of the desired item:
/// @code
///    auto& mybox = VBox      // vertical box
///    << Label("Settings")
///    << ( CheckItem("Bold")   << [&]{ setBold(); } )
///    << ( CheckItem("Italic") << [&]{ setItalic(); } )
///    << ( Button("Apply")     << [&]{ apply(); } );
/// @endcode
/// VBox create a vertical box, HBox an horizontal box. Boxes can be nested if needed:
/// @code
///    auto& mybox = VBox
///    << "Settings"s
///    << (HBox
///        << "Font"s
///        << ( TextField << [&]{ setFont(on::gadget()); } )
///        );
/// @endcode
/// The VBox contains a "Settings" Label and a HBox, which contains a "Font" Label and a TextField.
/// As in the previous examples "Settings" and "Font" are items. However, in a "simple" box such
/// as VBox and HBox they behave as Labels, so that their type can be omitted.
///
/// Note that "Settings" and  "Font" are followed by a 's' literal (which specifies they are C++ strings).
/// All strings can be postfixed by a 's', but this is only required when they are the first item of a box
/// (because of syntactical reasons).
///
///@section Attributed text
///
/// Text can be attributed using a pesudo-HTML syntax:
/// @code
///    Button("<b>Please <color=red>Click</color> <i>Here")
/// @endcode
/// The title of the Button will be in bold font, "Click" will be red and "Here" in italic font.
/// The available tags are: <b> (bold); <i> (italic); <color>, <background>, <font>.
/// The final ending tags can be omitted.
///
/// @section Cascaded Styles
///
/// Gadgets can refer to CSS-like styles:
/// @code
///    auto& mylist = ListBox("#mylist")
///    << [&]{ doIt(on::item()); }
///    << ".left"
///    << ".center"
///    << ".right";
/// @endcode
/// Contrary to the previous examples, the strings start with a . or a #.  These names refer to CSS
/// styles that can be specified in cascaded files and/or as a text in the program, for instance:
/// @code
///    #mylist {font: bold; border: rounded}
///    .left {text: Gauche}
///    #mylist .center {text: Centre; color: red}
/// @endcode
/// #mylist is equivalent to a HTML id and .left to a pseudo-class. Style selector can be contextual
/// (3rd line). Contrary to HTML/CSS, Guit styles can specify text content ('text' property  in the two
/// last lines), which is convenient for localisation.
///
/// In addition, Guit styles can also specify the items contained in the boxes:
/// @code
///    auto& mylist = ListBox("#mylist")
///    << [&]{ doIt(on::item()); };
/// @endcode
/// corresponding style, which relies on the 'items' property:
/// @code
///    #mylist {items: Gauche, Centre, Droite; font: bold}
/// @endcode
/// The items can refer to another style or embed a style:
/// @code
///    .left {text: Gauche}
///    #mylist {items: .left, {text: Centre; color:red}, Droite; font: bold}
/// @endcode
/// Styles can inherit from another style:
/// @code
///    .mypanel {font: bold; border: rounded}
///    #mylist { : .mypanel; items: Gauche, Centre, Droite}
/// @endcode
/// Finally, gagdets can refer several styles and contain inline styles:
/// @code
///    auto& mylist = ListBox(".mypanel, #mylist {color: red}")
///    << [&]{ doIt(on::item()); };
/// @endcode
///
/// @section Changing Properties
///
/// Gadgets' properties can be changed by using the << operator (as already seen), or by directly
/// changing the value of the properties. For instance, the text and the color of a TextField can
/// be changed as follows (first solution):
/// @code
///    auto& edit = TextField();
///    mytext << "Blah blah" << Color("blue");
///    mytext << "Whatever" << Color("red");
/// @endcode
/// Or (second solution):
/// @code
///    auto& text = Text();
///    auto& color = Color();
///    auto& edit = TextField() << text << color;
///
///    text = "Blah blah"; color = "blue";
///    text = "Whatever";  color = "red";
/// @endcode
/// The second solution is more efficient when there are many changes. It also allows syncing
/// several gadgets, as explained in the next section. Moreover, it provides a way to separate
/// the data and the logic from the gadgets, in a MVC style:
/// @code
///    // in header
///    struct M {
///       GText text;
///       GColor color;
///       GEnable enable;
///    } m;
///
///    // in source code
///    auto& tf = TextField() << text << color << enable;
/// @endcode
/// The logic of the program can then be implemented by managing the m variable, without having
/// to deal directly with gadgets. This is particularly useful when dealing with Tables (see GTable).
///
/// The Enable property controls whether a gadget is enabled (its value is true or false). Similarly,
/// the Show property controls if its is shown and the Select property if it is selected.
///
/// In this example the variables are declared as plain subobjects of a struct, which explains the
/// different syntax. For instance, GText is the class of the Text property, and Text() instantiates GText
/// (it returns *new GText). Classes always start with a G, intantiation functions don't.
/// See further section on about using plain subobjects.
///
/// @section Sharing Properties
///
/// Gadgets can be easily synced by sharing properties:
/// @code
///    auto& val = Float();
///    auto& color = Color();
///
///    auto& label  = Label() << val << color;
///    auto& edit   = TextField() << val << color;
///    auto& slider = Slider() << val;
/// @endcode
/// Changing the value of 'color' will then change the color of the Label and the TextField, and
/// changing the value of 'val' will update all three gadgets accordingly (the Label and the TextField
/// will show this value and the slider will be moved to the corresponding position). val can be
/// changed programmatically, or interactively by editing the TextField or moving the Slider.
///
/// This example also hightlights that a numeric prop (a Float or an Int) can not only be added
/// to a Slider, but also to a TextField or a Label (or any other item), altough they display text.
/// Moreover, the user will be forced to enter a float in the TextField
///
/// This feature can also serve for syncing boxes that perform exclusive selection, for instance
/// a box displaying tabs, and a StackBox, which allows displaying only one of its children:
/// @code
///    auto& choice = Choice(0);
///
///    auto& tabs = ChoiceBox
///    << choice
///    << "Home" << "Insert" << "Create" << "Animate";
///
///    auto& pages = StackBox
///    << choice
///    << home_page << insert_page << create_page << animate_page;
/// @endcode
/// 'home_page', etc. are boxes containing various controls. Because the 'choice' prop (which controls
/// the selection) is shared by 'tabs' and 'pages', their selection is synced, meaning that clicking
/// on the "Home" item will show 'home_page', on "Insert"  'insert_page', etc.
///
/// Note that Guit also provides a TabBox gadget for the same purpose. However the previous
/// solution is more general as it allows syncing various boxes in a uniform way. Note also that
/// adding a Choice property to any Box makes it enforce exclusive selection (for instance a Menu
/// can be made exclusive that way).
///
/// @section Binding Properties
///
/// Properties can be bound together using the bind() and sync() functions (or, equivalently,
/// the <<= and ^= operators):
///@code
///   auto& txt = Text();
///   auto& num = Float();
///   bind(txt, num);    // txt is bound to num
///   sync(txt, num);    // txt and num are synced
/// @endcode
/// - bind(txt, num) means that txt = num is performed whenever the value of num is changed
/// - sync(txt, num) bounds both way. Moreover, it forces its first argument to have a value
///  compatible with its second argument: txt will always represent a float, and if it is added
///  to a TextField the user will be forced to enter a float
///
///
/// example avec Enable
///
///
///
/// @section Active Expressions
///
/// Active Expressions are formulas that are automatically updated when the value of one
/// of their variables is changed:
/// @code
///    auto& euros = Float(100);
///    auto& dollars = Float();
///    auto& rate = Float(1.2);    // Euro => Dollar conversion rate
///
///    dollars <<= ~euros * ~rate;   // dollars depends on euros
///    euros <<= ~dollars / ~rate;   // euros depends on dollars
/// @endcode
/// In this example, 'dollars' is auto-updated when the value 'euros' or 'rate' is changed
/// (and vice-versa for 'euros'). The <<= operator has the same effect as bind().
///
/// Note that variables must be prefixed by a ~ in Active Expressions. This avoids syntactical
/// ambiguities by indicating which variables auto-update the expression.
///
/// Active expressions can also be injected in the GUI specification, e.g. to specify the text
/// a gadget will display:
/// @code
///     auto& box = VBox
///     << (TextField << euros)
///     << (TextField << rate)
///     << ("Result: " + (~euros * ~rate) + "$");
/// @endcode
///
/// @section Conditional Expressions and Callback Functions
///
/// The syntax is: trigger / action, which means that the action
/// will be performed when the trigger if fired:
/// @code
///   auto& mybutton = Button("Do It")
///   << [&]{ doIt(); }
///   << on::action / [this]{doIt();}  // equivalent to previous line
///   << on::mousedown / [&]{ doMouseDown(); }
///   << on::mouseup   / [&]{ doMouseUp(); };
/// @endcode
/// A trigger can be an Event Type (see On class) or an Active Expression
/// (see below). When the trigger is ommitted, a default trigger is used.
/// When this trigger is fired depends on which gadget it is added to:
///
/// In non-box gadgets:
/// - When the gadget is activated (ie clicked) if it has no state (e.g. an
///   Item is a Menu)
/// - When the gadget is selected or deselected if it has a state (e.g. a
///   CheckItem in a menu)
/// - When the gadget is selected if its has a state and is part of Box that
///   enforces exclusive selection (e.g. an Item in a ListBox or a RadioBox)
/// - etc. (see gadget classes)
///
/// In box gadgets:
/// - When the default trigger of the box children is fired. For instance,
///   in the first
///   example, the Menu callback is be fired when one of its items is activated.
///
/// See GTrigger class for for more detail. Note that the on::action Event Type denotes the default trigger
/// (specifying on::action or omitting the trigger does the same thing).
///
///
/// @section Conditional Props
///
/// Properties can be conditionally added to gadgets using a similar syntax:
/// @code
///   auto& btn = Button("Select Me")
///   << on::select   / Color("red")
///   << on::deselect / Color("blue");
/// @endcode
/// The color of the button then depends on whether it is selected or not:
/// The red color is added when the button selected, and the blue color when
/// it is deselected. As a gadget can have only one color, the red color
/// replaces the blue color and vice-versa.
///
/// @code
///     auto& status = Int(0);  // integer prop
///     auto& label = Label
///     << (~status == 0) / "<color=green>OK"
///     << (~status == 1) / "<color=orange>Warning"
///     << (~status == 2) / "<color=red>DANGER!!!"
///
///     status = 2;
/// @endcode
///
/// @section Layout
/// Children are auto-laid out except if they have a GPos.
/// These props allow specifying a (x,y) position and (optionally) a layer,
/// which controls how they are superimposed.
///
/// Setting the position:<br>
/// The position is computed automatically except if a GPos is added to
/// the gadget, in which case it is displayed on the top of its siblings.
/// A GPos can also specify a layer.
/// @see GPos, x(), y().
///
/// Setting the size:<br>
/// The size is computed automatically except if a GSize is added to
/// the gadget. A GSize can specify a preferred size, a fixed size,
/// and how the gadget can be resized by its parent. The final size also
/// depends on the GAlign prop of the parent.
/// @see GSize and GAlign.
///
/// @section Automatic Deletion and Smart Pointers
///
/// Gadgets and properties are auto-deleted when they are removed from (all) their parents.
/// Thus C++ raw pointers and references may become pending and should only be used
/// for referencing objects temporarily.
///
/// gptr<> smart pointers provide a way to "own" objects: they won't be deleted if at least
/// one gptr points to them:
/// @code
///    gptr<GChoice> choice = Choice(2);
///    gptr<GText> text = Text();
///
///    gptr<GComboBox> combo = ComboBox
///    << *choice
///    << *text
///    << "Left" << "Center" << "Right";
/// @endcode
/// Note that gptrs must be deferenced, e.g.:  ComboBox << *choice << *text
///
/// shared_ptr<> and unique_ptr<> are not compatible with Guit objects.
/// The delete operator should not be used (it does nothing except issuing a warning).
/// To delete an object safely do:
/// @code
///    gptr<GObject> ptr = obj;  // whatever object to delete
///    ptr = nullptr;
/// @endcode
/// This will delete the object pointed by 'obj' if it's not pointed elsewhere (by a gptr
/// or a parent gadget)
///
/// @section Plain Objects and SubObjects
/// Plain objects can be used provided that auto-deletion is disabled on these objects.
/// The simplest solution consistsin using gplain<> :
/// @code
///    struct M {
///       GText text;
///       GColor color;
///       GEnable enable;
///       gplain p{text, color, enable};   // disables auto-deletion
///    } m;
///
///    class GUI {
///       GFloat value;
///       GText mytext;
///       M m;
///
///       GUI() {
///         gplain{value, mytext};  // disables auto-deletion
///        // etc.
///       }
///    };
/// @endcode
/// gplain<> can either be used as object (first case) or as a function (second case).
/// Alternately, the setPlain() method can be called on each plain object.
///
/// @section Implementation
/// Different gadgets with different types may have the same
/// implementation class, but
/// they initialize them differently. For instance  Item, Label, Button,
/// CheckItem  all return an object of the GItem class. Class<Item>,
/// Class<Label>, etc. thus all denote the same class (i.e. GItem), but
/// Class<Box> is another class (i.e. GBox). Class<> avoids searching
/// which is the actual implementation class.
