//
//  MenuBar, ToolBar, StatusBar
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Bar_hpp
#define Guit_Bar_hpp
#include <gbox.hpp>
namespace guit {

/// Menu bar.
/// When added to a Box a MenuBar gadget is placed in the top part of the Box' border.
/// Strings added to this gadget are converted into MenuBar Items (see GItem). As for any Item,
/// a Menu (see GMenu) can for instance be added to them. Pressing the Item will open this Menu.
/// @code
///   auto& bar = MenuBar
///   << ("File"
///        << (Menu
///             << ("Open" << [this]{doOpen();})
///             << ("Save" << [this]{doSave();})
///           )
///      )
///   << ("Edit"
///        << (Menu
///             << ("Undo" << [this]{doUndo();})
///             << ("Redo" << [this]{doRedo();})
///           )
///      );
/// @endcode
/// @see GBox and Gadget.
class GMenuBar : public GButtonBox {
protected:
  GMenuBar(GadgetType&, GString const& css_and_title);

public:
  GMenuBar(GString const& css_and_title = "");

  GMenuBar* toMenuBar() override {return this;}

  using GButtonBox::add;
  void add(GMenu* child) override;

  Place place() const override {return Place::Top;}

  using GType = GadgetType_<GMenuBar,GButtonBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GMenuBar& MenuBar(GString const& css_and_title = "");



/// Tool bar.
/// When added to a Box a ToolBar is placed in the top part of the Box' border, below the
/// MenuBar if there is one
/// Strings added to this gadget are converted into ToolBar Items (see GItem). As for any Item,
/// a callback or conditional expression (see GCond) can for instance be added to them.
/// @code
///    auto& bar = ToolBar
///    << ("Open" << [this]{doOpen();})   // creates a ToolBar item
///    << ("Save" << [this]{doSave();});
/// @endcode
/// @see GBox and Gadget.
class GToolBar : public GButtonBox {
protected:
  GToolBar(GadgetType&, GString const& css_and_title);

public:
  GToolBar(GString const& css_and_title = "");

  Place place() const override {return Place::ToolBar;}

  using GType = GadgetType_<GToolBar,GButtonBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GToolBar& ToolBar(GString const& css_and_title = "");



/// Status bar.
/// When added to a Box a StatusBar is placed in the bottom part of the Box' border.
/// Strings added to this gadget are converted into StatusBar items.
/// @code
///    auto& status = Text();
///    auto& sbar = StatusBar << (Label << status);
///    status << "My Message";    // changes the status
/// @endcode
/// @see GBox and Gadget.
class GStatusBar : public GBox {
protected:
  GStatusBar(GadgetType&, GString const& css_and_title);

public:
  GStatusBar(GString const& css_and_title = "");

  Place place() const override {return Place::Bottom;}

  using GType = GadgetType_<GStatusBar,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  Gadget& createDefaultItem(GString const& arg) override;
};

GStatusBar& StatusBar(GString const& css_and_title = "");

}
#endif
