//
// Item Gadgets: labels, buttons and menu/list items.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gbind.hpp>
#include <gevent.hpp>
#include <gborder.hpp>
#include <gitem.hpp>
#include <gmenu.hpp>
#include <gtext.hpp>
#include <gicon.hpp>
#include <gmodes.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(Item)
GIMPLEMENT_GADGET_ALL(Separator, GItem)
GIMPLEMENT_GADGET_ALL(Filler, GItem)
GIMPLEMENT_GADGET_ALL(Label, GItem)
GIMPLEMENT_GADGET_ALL(Button, GItem)
GIMPLEMENT_GADGET_ALL(LinkItem, GItem)
GIMPLEMENT_GADGET_ALL(ChoiceItem, GItem)
GIMPLEMENT_GADGET_ALL(CheckItem, GItem)
GIMPLEMENT_GADGET_ALL(RadioItem, GItem)
GIMPLEMENT_GADGET_ALL(Title, GItem)

// ambigu bec GFunction
//  GItem& operator<<(GString const& item, gptr<GProp> prop) {
//  return *new GItem(item) << prop;
//}
/// Creates an item, adds a prop, a menu, a window/dialog or a function to it, returns the item.
/// - When a menu is added to the item, pressing the item opens the menu
/// - When a window/dialog is added to the item, clicking on the item opens the window/dialog

GItem::~GItem() {
  delete tcache_;
}

GItem::GItem(GadgetType& t, GString const& spec)
: Gadget(t, spec) {}

GItem::GItem(GString const& spec)
: Gadget(Type(), spec) {}

GItem& GItem::clone(GClone const& c) {
  auto* g = new GItem;
  g->initClone(*this, c);
  return *g;
}

GItem& GItem::setMultiLine(bool value) {
  modes_.multiLine = value;
  update();
  return *this;
}

GItem& GItem::operator=(GString const& s) {
  prop<GText>() = s;
  return *this;
}

GItem& GItem::operator=(float val) {
  GString str;
  gconvert(val, str);
  return operator=(str);
}

GString GItem::text() const {
  auto* t = findProp<GText>();
  if (t) return t->value();
  else if (style_ && (t = style_->findProp<GText>(*this))) return t->value();
  else return emptystring_;
}

GString GItem::stringValue() const {
  return isChecked() ? "true" : "false";
}

GProp* GItem::valueProp() {return findProp<GCheck>();}

void GItem::setValue(GString const& text) {
  bool state = false;
  gconvert(text, state);
  check(state);
}

void GItem::add(GMenu* opened) {
  if (!opened) error("add","Can't add null child");
  // mousedown, any button, autoplace
  else addOpener(*opened, GTrigger::mousedown, -1, true);
}

void GItem::add(GWindow* opened) {
  if (!opened) error("add","Can't add null child");
  // action, button 1, autoplace
  else addOpener(*opened, GTrigger::action, 1, true);
}

void GItem::addProp(GProp& prop, bool add_to_list) {
  switch (prop.role()) {

    case GPropType::Bool:
      if (!prop.isMutable()) check(prop.boolValue());
      else if (auto* p = prop.toBool()) {
        addProp2(prop, add_to_list);
        gsync(this->prop<GCheck>(), *p, -1);
      }
      updateLayout();
      return;

    default:
      Gadget::addProp(prop, add_to_list);
      return;
  }
}

void GItem::doAction() {
  GPoint pos(frame_.x, frame_.y);
  pos.x += w() / 2;
  pos.y += h() / 2;
  GMouseEvent e(on::action, GCore::elapsedTime());
  e.button_ = 1;
  e.wpos_ = pos;
  doFire(e);
}

void GItem::pack(GRender* parr) {
  if (!isMultiLine()) return;

  GBox* par = getParent();
  if (!par) return;

  auto* text = findProp<GText>();
  if (!text) return;

  if (!tcache_) tcache_ = new GTextCache;
  //if (rowFlexRef(this) == 0 || columnFlexRef(this) == 0) return;

  GRender r(*this, parr);
  auto dim = GTextRender::layoutText(r.font_, modes_.richText, text->value(),
                                     par->childrenArea(*parr).w,
                                     nullptr, tcache_);
  if (r.margin_)
    frame_.h = dim.h + r.margin_->top + r.margin_->bottom;
  else
    frame_.h = dim.h;
}

}
