//
//  Item Gadgets with a numerical value
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gtrigger.hpp>
#include <gevent.hpp>
#include <gbind.hpp>
#include <gint.hpp>
#include <gfloat.hpp>
#include <gvalueitem.hpp>
#include <gtext.hpp>
#include <gfloat.hpp>
#include <gbox.hpp>
#include <gcontrol.hpp>
#include <gadgetstyle.hpp>
#include <grender.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(ValueItem)

GValueItem::~GValueItem() {}

GValueItem::GValueItem(GadgetType& t, GString const& spec) : GItem(t),
value_(new GFloat) {
  addProp(*value_, true);
  *this << on::mousedrag / [this](GEvent&e){mouseDrag(e.asMouseEvent());};

  if (!spec.empty()) initSpec(spec);
}

void GValueItem::initSpec(GString const& spec) {
  GString title;
  GadgetStyle::initStyle(*this, spec, title);

  if (!title.empty()) {
    if (!::isdigit(title[0])) add(title);
    else {
      GString val;
      size_t from{};
      strtoken(title, val, from);
      value_->set(val);

      if (from >= title.size()) return;
      if (::isspace(title[from])) from++;
      if (from >= title.size()) return;
      add(title.substr(from));
    }
  }
}

GValueItem::GValueItem(GString const& spec)
: GValueItem(Type(), spec) {}

GValueItem::GValueItem(float value)
: GValueItem(Type(),"") {setValue(value);}

GValueItem& ValueItem(float val) {return *new GValueItem(val);}

GValueItem& GValueItem::clone(GClone const& cm) {
  auto* g = new GValueItem;
  g->initClone(*this, cm);
  return *g;
}

void GValueItem::initClone(Gadget& from, GClone const& c) {
  GItem::initClone(from, c);
  if (c & GClone::ShareProps) {
    if (auto* f = from.toValueItem()) value_ = f->value_;
  }
}

GNumber& GValueItem::operator()() {
  return *value_;
}

float GValueItem::value() const {
  return value_->floatValue();
}

GString GValueItem::stringValue() const {
  GString txt;
  gconvert(value(), txt);
  return txt;
}

void GValueItem::setValue(float value) {
  value_->setF(value);
}

void GValueItem::setValue(GString const& value) {
  float v{};
  if (gconvert(value, v)) value_->setF(v);
  else error("setValue","value should be a number: "+value);
}

float GValueItem::min() const {return value_->minF();}
float GValueItem::max() const {return value_->maxF();}

GValueItem& GValueItem::setRange(float min, float max) {
  value_->setRangeF(min, max);
  return *this;
}

void GValueItem::addProp(GProp& prop, bool add_to_list) {
  switch (prop.role()) {

    case GPropType::Number:
    case GPropType::Choice:

      if (!prop.isMutable()) {
        setValue(prop.floatValue());
      }
      else if (GNumber* num = prop.toNumber()) {
        if (auto* f = num->toFloat()) {
          value_ = f;
          num->startNotify(*this, [this](auto&){updateLayout();});
        }
        else if (auto* i = num->toInt()) {
          gsync(*value_, *i);
        }
        else {
          error("addProp","A "+prop.typeName()+" prop can't be added to this gadget");
          return;
        }

        addProp2(prop, add_to_list);  // was above

        // initially in GSlider; othewise the valueitem wont appear in a box
        // if range not set, set it to [0,100]
        float min = num->minF(), max = num->maxF();
        if (min == num->lowestF()) min = 0.f;
        if (max == num->highestF()) max = 100.f;
        value_->setRangeF(min, max);
        update();
      }
      return;

    default:
      GItem::addProp(prop, add_to_list);
      return;
  }
}

GFlow::Kind GValueItem::getFlow() {
  auto* par = getParent();
  if (par) {
    if (par->isPie()) return GFlow::Pie;
    else if (par->isRow()) return GFlow::Column;
    else return GFlow::Row;
  }
  else return GFlow::Row;
}

void GValueItem::movedResizedCB(GRect const& initframe) {
  switch (getFlow()) {
    case GFlow::Row:
      offset_ = float(frame_.w * value_->normedValue());
      break;
    case GFlow::Column:
      offset_ = float(frame_.h * value_->normedValue());
      break;
    default:
      break;
  }
  GItem::movedResizedCB(initframe);
}

void GValueItem::mouseDrag(GMouseEvent& e) {
  if (!isTrulyEnabled()) return;  // || !isEdited

  switch (getFlow()) {
    case GFlow::Row:
      value_->increaseF(e.dx() * (max() - min()) / frame_.w);
      return;

    case GFlow::Column:
      value_->increaseF(-e.dy() * (max() - min()) / frame_.h);
      return;

    case GFlow::Pie:
      value_->increaseF(e.dx() * (max() - min()) / frame_.w
                        -e.dy() * (max() - min()) / frame_.h);
      return;

    default:
      return;
  }
}

void GValueItem::doPaint(GRenderGraphics& gr, GRender* parr) {
  GRender r(*this, parr);
  auto dir = getFlow();

  if (dir == GFlow::Pie) {  // Pie display
    if (!r.paintGadget(gr, GRender::DrawBackground)) return;
  }
  else {  // Row or Column display
    if (!r.paintGadget(gr, GRender::DrawBorder)) return;
    gr.setBgColor(r.background_);
    gr.unsetColor();
    paintContent(gr, r, dir);
  }
  if (r.text_) r.drawText(gr, offset_);
  if (isChecked() && layout_.reshapeFrame) paintHandleFrame(gr);
}

void GValueItem::paintContent(GRenderGraphics& gr, GRender& r, GFlow::Kind dir) {
  auto area = frame_;
  switch (dir) {
    case GFlow::Row:
      area.w = offset_;
      break;
    case GFlow::Column:
      area.y += area.h - offset_;
      area.h = offset_;
      break;
    default:
      break;
  }
  //if (b.style == b.Rounded) gr.drawRoundRect(area, 5.,5.);
  //else gr.drawRect(area);
  gr.drawRect(area);

  if (isChecked()) {
    float d = 4;
    if (d+d >= area.w-10 || d+d > area.h-10) {d = 2;}
    area.x += d;
    area.y += d;
    area.w -= d+d;
    area.h -= d+d;
    r.setAutoColor(gr);
    gr.unsetBgColor();
    gr.setThickness(2);
    //if (b.style == b.Rounded) gr.drawRoundRect(area, 5.,5.);
    //else gr.drawRect(area);
    gr.drawRect(area);
  }
}

}
