//
//  Mouse Cursor
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Cursor_hpp
#define Guit_Cursor_hpp
#include <gprop.hpp>
#include <gproplist.hpp>
namespace guit {

class GPos;

/// System or Pseudo Cursor.
class GCursor : public GProp {
public:
  using Shape = uint32_t;
  
  static const Shape
  Arrow,
  Crosshair,
  IBeam,
  Wait,
  WaitArrow,
  Resize_W_E,
  Resize_N_S,
  Resize_NE_SW,
  Resize_NW_SE,
  ResizeAll,
  NoAction,
  Hand;

  //static GPropType_<GCursor>& Type();
  //GPropType& type() const override {return Type();}

  GCursor* clone(bool) const override {return nullptr;}

  /// Returns the size of the main screen.
  static GDim screenSize();
  
  /// Returns the cursor corresponding to this input channel if it exists, null otherwise.
  /// @see obtainCursor().
  static GCursor* cursor(unsigned int channel = 0);

  /// Returns the cursor corresponding to this channel, creates it if needed.
  /// - 0 specifies the system cursor, it is linked to the native event flow.
  /// - strictly positive values specify pseudo-cursors. Each pseudo-cursor is linked
  ///  to the event flow of a separate input channel.
  /// Pseudo-cursors can only be displayed on Guit windows (not on the entire screen)
  /// @see cursor()
  static GCursor& obtainCursor(unsigned int channel = 0);

  /// Returns the number of cursors that currently exist.
  /// Always >= 1 because this count includes the system cursor.
  static int cursorCount();

  /// returns the channel of this cursor.
  unsigned int channel() const {return id_;}

  /// Moves the cursor on the screen (screen coordinates).
  /// @{
  void move(GPoint const& pos);
  void move(float x, float y) {move(GPoint(x,y));}
  /// @}

  /// Returns the position of the cursor on the screen.
  /// @{
  GPoint const& pos() const;
  float x() const {return pos().x;}
  float y() const {return pos().y;}
  /// @}

  /// Returns the position of the cursor relative to the current window.
  /// Return {0,0} if no current window
  /// @{
  GPoint wpos() const;
  float wx() const {return wpos().x;}
  float wy() const {return wpos().y;}
  /// @}

  /// Adds a callback that is fired when the mouse value is changed.
  GCursor& operator<<(GFunction const& fun);

  /// binds this pos with the cursor.
  friend GPos& operator<<=(GPos& pos, GCursor&);

  /// Shows/hides this cursor.
  void show(bool state);

  /// Freezes this cursor.
  /// The cursors won't move if state is true. However if will still receive events.
  void freeze(bool state);

  /// Changes the color of the cursor.
  void setColor(GColor&);

  /// Returns the color of the cursor.
  GColor& color() const {return *color_;}
  
  /// Changes the shape of the cursor.
  /// Works only if this cursor display a native cursor (typically, cursor(0)).
  void setShape(Shape);

  /// Returns the shape of the cursor.
  Shape shape() const {return shape_;}

  /// Displays this cursor as a native cursor.
  /// This cursor will be displayed as a native cursor, even if it's not the
  /// system cursor. This feature allows moving the native cursor in a customized
  /// way that don't depend on the native event flow, e.g. to change the mouse
  /// gain dynamically or to process input from multiple touchpads differently.
  /// The system cursor (cursor(0)) should then typically be frozen, meaning
  /// it will still receive events but won't move the cursor on the screen.
  /// This method has no effect of the system cursor (cursor(0)).
  void displayAsNative(bool);

  /// Starts a Drag & Drop action controlled by this cursor.
  /// _onDrop_ is called when the drag-and-drop operation ends.
  void startDrag(Gadget& from,
                 std::function<void(GCursor&)> const& onDrop,
                 Shape = Hand);
  
  /// Returns true if a Drag & Drop is being performed.
  bool isDragging() const;
  
  /// Returns the gadget where the Drag & Drop started from (if any).
  Gadget* source();
  
  /// Returns the window where the Drag & Drop started from (if any).
  GWindow* sourceWindow();
  
  /// Returns the gadget where the Drag & Drop ended (if any).
  Gadget* target() {return target_;}
  
  /// Returns the window where the Drag & Drop ended (if any).
  GWindow* targetWindow() {return targetwin_;}
  
  /// Returns the position in the window where the Drag & Drop ended.
  GPoint const& targetPos() {return targetpos_;}

protected:
  friend class GChannel;
  friend class GControlImpl;

  GCursor(unsigned int channel_id);
  void channelDown(GMouseEvent&);
  void channelUp(GMouseEvent&);
  void channelMove(GMouseEvent&);
  static void drawCursors(GRenderGraphics&);
  void reset();
  void endDrag(GPoint const& wpos);
  void moveDrag(GPoint const& wpos);

public:
  unsigned int id_{};
  bool shown_{}, pressed_{}, dragging_{}, nativecursor_{};
  int32_t derive_{}, maxderive_{}, derivecount_{};  // NOT signed!
  Shape shape_{Arrow};
  gprotect<GColor> color_;
  gprotect<GWindow> window_;
  GPropList props_;
  mutable GPoint pos_;
  GPoint targetpos_;
  GWindow *fromwin_{}, *targetwin_{};
  Gadget *fromg_{}, *target_{};
  std::function<void(GCursor&)> onend_;
  std::function<bool(GMouseEvent&)> buttonhandler_{};
  std::function<bool(GMouseEvent&)> movehandler_{};
  std::function<bool(GMouseEvent&)> wheelhandler_{};
  std::function<bool(GKeyEvent&)>   keyhandler_{};
};

}
#endif
