//
//  Events
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitEvent_hpp
#define GuitEvent_hpp
#include <gtypes.hpp>
namespace guit {

class GTrigger;
class GMouseEvent;
class GKeyEvent;
class GTouchEvent;
class GGestureEvent;

/// Event types.
/// @{
enum struct GEventType : uint32_t {
  NoEvent     = 0,
  Action      = 1<<0,
  Done        = 1<<1,
  Click       = 1<<2,
  Change      = 1<<3,
  Changing    = 1<<4,
  MouseButton = 1<<5,
  MouseDrag   = 1<<6,
  MouseMove   = 1<<7,
  MouseCross  = 1<<8,
  Wheel       = 1<<9,
  Key         = 1<<10,
  Caret       = 1<<11,
  Touch       = 1<<12,
  Gesture     = 1<<13,
  Focus       = 1<<14,
  Enable      = 1<<15,
  Show        = 1<<16,
  Move        = 1<<17,
  Paint       = 1<<18,
  Resize      = 1<<19,
  Init        = 1<<20,
  AddRemove   = 1<<21,
  CutPaste    = 1<<22,
  DragDrop    = 1<<23,
  User        = 1<<30
};
inline GEventType operator|(GEventType a, GEventType b) {
  return GEventType(uint32_t(a) | uint32_t(b));
}
inline GEventType operator&(GEventType a, GEventType b) {
  return GEventType(uint32_t(a) & uint32_t(b));
}
inline bool operator==(GEventType a, int b) {return uint32_t(a) == uint32_t(b);}
inline bool operator!=(GEventType a, int b) {return uint32_t(a) != uint32_t(b);}
/// @}


/// Button and Modifier states.
/// GMouseEvent and GKeyEvent have a state() method that returns and ORed combination of these states.
/// Button1 corresponds tothe  left mouse button, Button2 to the right button, Button3 to the middle button.
/// @{
enum struct GEventState : uint32_t {
  NoState = 0,
  Button1 = 1<<1,
  Button2 = 1<<2,
  Button3 = 1<<3,
  Button4 = 1<<4,
  Button5 = 1<<5,
  Ctrl    = 1<<20,
  Shift   = 1<<21,
  Alt     = 1<<22,
  AltGr   = 1<<23,
  Meta    = 1<<24,
  Caps    = 1<<25,
  Num     = 1<<26
};
inline GEventState operator|(GEventState a, GEventState b) {
  return GEventState(uint32_t(a) | uint32_t(b));
}
inline GEventState operator&(GEventState a, GEventState b) {
  return GEventState(uint32_t(a) & uint32_t(b));
}
inline bool operator==(GEventState a, int b) {return uint32_t(a) == uint32_t(b);}
inline bool operator!=(GEventState a, int b) {return uint32_t(a) != uint32_t(b);}
/// @}


/// GUI Event: base of all events.
/// GUI events are produced when a trigger is fired, see GTrigger.
/// @see subclasses GMouseEvent, GKeyEvent, etc.
class GEvent {
public:

  /// Keycodes of special keys.
  /// GKeyEvent has a key() method that returns the keycode.
  static const GKeyCode
  EscapeKey, TabKey, BackspaceKey, SpaceKey, ReturnKey, EnterKey, DeleteKey,
  LeftKey, RightKey, UpKey, DownKey, HomeKey, EndKey,
  CtrlKey, ShiftKey, AltKey, AltGrKey, MetaKey, CapsKey, NumKey;


  GEvent(GTrigger const& trigger, GTime time);
  virtual ~GEvent();

  /// Returns the event casted to the specified event subclass.
  /// Throws an exception that resets the event loop if the event does not derive from this subclass.
  /// @{
  virtual GMouseEvent& asMouseEvent();
  virtual GKeyEvent&   asKeyEvent();
  virtual GTouchEvent& asTouchEvent();
  virtual GGestureEvent& asGestureEvent();
  /// @}

  /// Returns the event casted to the specified event subclass.
  /// Returns null if if the event does not derive from this subclass.
  /// @{
  virtual GMouseEvent* toMouseEvent() {return nullptr;}
  virtual GKeyEvent*   toKeyEvent()   {return nullptr;}
  virtual GTouchEvent* toTouchEvent() {return nullptr;}
  virtual GGestureEvent* toGestureEvent()  {return nullptr;}
  /// @}

  /// Returns where this event occured.
  /// - If the source is a Gadget, from() returns this gadget
  /// - If the source is a GBox gadget, fromBox() returns from() casted to GBox
  /// - If the source is a GProp, fromProp() returns this prop
  /// A GFatalError exception is thrown if the source is not from the expected type, which will
  /// quit all callback functions and restart the event loop.
  /// @code
  ///   auto& txt = TextField()
  ///   << [](GEvent& e){ std::cout << e.from().text(); };
  /// @endcode
  /// @{
  Gadget& from() const;
  GBox&   fromBox() const;
  GProp&  fromProp() const;
  /// @}

  /// Returns the checked child of a box that enforces exclusive selection.
  /// Returns null if N/A, and the same value as  GChoiceBox::chosen() otherwise.
  /// @code
  ///   auto& list = ListBox()
  ///   << "one"
  ///   << "two"
  ///   << "three"
  ///   << [](GEvent& e){ if (e.chosen()) std::cout << e.chosen()->text(); };
  /// @endcode
  Gadget* chosen() const;

  /// Returns the window where this event occured.
  GWindow* window() const;

  /// Returns the trigger that fired this event.
  GTrigger const& trigger() const {return *trigger_;}

  /// Returns the event mask.
  GEventType eventMask() const;

  /// Returns the input channel.
  /// Return 0 for system events, a value > 0 for other input channels (see GChannel).
  int channel() const {return channel_;}

  /// Returns the timestamp.
  /// The timestamp is only defined for mouse/finger/key events.
  GTime time() const {return time_;}

  /// Sets a function that is called after firing all callbacks trigerred by a same event.
  void postFire(std::function<void(GEvent&)> const& todo_after);

  /// Changes/returns the event tag.
  /// Can be useful when redirecting events (see GMouseEvent::redirectEvent) so that the gadget
  /// that redirects the event can pass data to the final receiver.
  /// @{
  GObject* tag() const {return tag_;}
  void setTag(GObject* tag) {tag_ = tag;}
  /// @}

  /// Returns a non-zero value if this trigger matches this event.
  int matches(GTrigger const& trigger, int channel) const;

  /// Sets a global handler that intercepts all the the events of the specified type.
  /// Events will then processed normally if the handler returns true, and
  /// ignored otherwise. Only one handler can be set for a given event type.
  /// Passing nullptr removes the corresponding handler.
  /// @{
  static void setMouseButtonHandler(int channel,std::function<bool(class GMouseEvent&)>);
  static void setMouseMotionHandler(int channel, std::function<bool(class GMouseEvent&)>);
  static void setWheelHandler(int channel, std::function<bool(class GMouseEvent&)>);
  static void setKeyHandler(int channel, std::function<bool(class GKeyEvent&)>);
  /// @}

  /// Converts a key code into a key name and vice-versa
  /// @{
  static GString keyToName(GKeyCode key);
  static GKeyCode nameToKey(GString const& key_name);
  /// @}

  /// Converts a mouse button number into an event state.
  /// mouse_button = 1 returns GEventState::Button1 (left button), 2 returns Button 2 (right button),
  /// 3 returns Button3 (middle button), etc.
  static GEventState buttonToState(int mouse_button);

  /// Converts a modifier keycode intoto an event state.
  static GEventState modifierToState(GKeyCode modifier_key);

  /// Converts a combination of modifier names to ORed modifier states.
  /// Valid modifier names are: ctrl, shift, meta/cmd, alt, altgr, caps, num.
  /// They can be combined by adding a - or a + between them, e.g. ctrl-shift
  static GEventState modifiersToStates(GString const& modifiers);

  /// @internal
  /// Note: As Guit is a research toolkit, it provides full access to all fields,
  /// but beware that using them is at your own risk.
  /// @{
  bool hasGuard() const {return hasguard_;}
  bool guardDone() const {return guarddone_;}

  GTrigger const* trigger_{};
  bool hasguard_{}, guarddone_{}, reshapeinter_{};
  int channel_{};
  GWindow* window_{};
  Gadget*  gadget_{};
  GObject *aux_{}, *tag_{};
  GTime time_{};
  std::function<void(GEvent&)> postfire_{};
  /// @}
};


/// Mouse event: event when the mouse is pressed, released, dragged or moved.
/// Corresponding triggers: GTrigger::mousedown, mouseup, mousedrag, mousemove
class GMouseEvent : public GEvent {
public:
  GMouseEvent(GTrigger& trigger, GTime t) : GEvent(trigger, t) {}
  GMouseEvent(GTrigger& trigger, GMouseEvent const&);

  GMouseEvent& asMouseEvent() override {return *this;}
  GMouseEvent* toMouseEvent() override {return this;}

  /// Returns the position relative to the sender().
  /// @{
  GPoint pos() const;
  float x() const {return pos().x;}
  float y() const {return pos().y;}
  /// @}

  /// Returns the position relative to the window().
  /// @{
  const GPoint& wpos() const {return wpos_;}
  float wx() const {return wpos_.x;}
  float wy() const {return wpos_.y;}
  /// @}

  /// Returns the displacement from last GMouseEvent.
  /// @{
  float dx() const {return dx_;}
  float dy() const {return dy_;}
  /// @}

  /// Returns the button that was just pressed or released.
  /// Returns an int that is > 0 for *mousedown* and *mouseup* events (see GTrigger)
  /// and 0 otherwise. 1 indicates the left mouse button, 2 the right button,
  /// 3 the middle button, etc.
  /// @note Returns 0 for *mousemove* and *mousedrag* events.
  /// @note Use state() to get the buttons that were previously pressed (if any).
  int button() const {return button_;}

  /// Returns the number of mouse clicks on *click* events.
  /// The returned value should only be used for *click* events (see GTrigger).
  int clickCount() const {return clickcount_;}

  /// Returns an ORed combination of buttons and modifiers; see GEvent::States.
  /// @Note for on::mousedown events the button that was just pressed is not included in the state.
  GEventState state() const {return state_;}

  /// Returns the modifier keys that are currently pressed.
  /// @{
  bool shiftPressed() const {return (state_ & GEventState::Shift) != 0;}
  bool ctrlPressed()  const {return (state_ & GEventState::Ctrl) != 0;}
  bool metaPressed()  const {return (state_ & GEventState::Meta) != 0;}
  /// @}

  /// Redirects the event to this gadget.
  /// Useful when creating gadgets interactively: typically the event is fisrt caught by a drawing
  /// panel, a gadget can then be created on this panel and the event can redirected to this gadget
  /// as if it received the event. This allows resizing or moving the gadget interactively it it has
  /// a GDragSize or a GDragPos prop.
  void redirectEvent(Gadget& to_gadget);

  /// @internal
  /// @{
  int button_{}, clickcount_{};
  GEventState state_{};
  GPoint wpos_;
  float dx_{}, dy_{};
  /// @}
};


/// Key event:  event when a key is pressed/released or when text in entered.
/// Corresponding triggers: GTrigger::keydown, keyup, type.
class GKeyEvent : public GEvent {
public:
  GKeyEvent(GTrigger const& trigger, GTime t) : GEvent(trigger, t) {}

  GKeyEvent& asKeyEvent() override {return *this;}
  GKeyEvent* toKeyEvent() override {return this;}

  bool shiftMod() const {return (state_ & GEventState::Shift) != 0;}
  bool ctrlMod() const  {return (state_ & GEventState::Ctrl) != 0;}
  bool metaMod() const  {return (state_ & GEventState::Meta) != 0;}

  /// Returns an ORed combination of buttons and modifiers; see GEvent::States.
  GEventState state() const {return state_;}

  /// Returns the keycode of the key that was pressed or released.
  /// @Note may return 0 for on::keytype events.
  GKeyCode key() const {return key_;}

  /// Returns the entered text (in UTF8).
  /// @Note may return "" for on::keydown and on::keyup events or if a modifier was pressed
  /// (e.g. ctrl-K)
  GString const& text() const {return text_;}

  /// @internal
  /// @{
  GEventState state_{};
  GKeyCode key_{};
  GString text_;
  /// @}
};


/// Touch event: event when a finger touches the touchpad.
/// Note that the sender is always the Main Window (GApp::mainWindow()).
/// Corresponding trigger: GTrigger::touch
class GTouchEvent : public GEvent {
public:
  enum Type {Down, Up, Motion};

  GTouchEvent(GTrigger const& trigger, GTime t) : GEvent(trigger, t) {}

  GTouchEvent& asTouchEvent() override {return *this;}
  GTouchEvent* toTouchEvent() override {return this;}

  Type type_{};
  int64_t touchID_{}, fingerID_{};
  GPoint pos_;        // normalized in [0.,1.]
  GPoint dist_;       // normalized in [0.,1.]
  float pressure_{};  // normalized in [0.,1.]
};


/// Gesture event: event when a gesture is recognized.
/// Note that the sender is always the Main Window (GApp::mainWindow()).
/// Corresponding trigger: GTrigger::gesture
class GGestureEvent : public GEvent {
public:
  GGestureEvent(GTrigger const& trigger, GTime t) : GEvent(trigger, t) {}

  GGestureEvent& asGestureEvent() override {return *this;}
  GGestureEvent* toGestureEvent() override {return this;}

  int64_t touchID_{};
  int numfingers_{};
  GPoint pos_;   // normalized in [0.,1.]
  float dist_{}, theta_{};
};

}

#endif
