//
//  Smart Pointers.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef guit_ptr_hpp
#define guit_ptr_hpp
#include <gobject.hpp>
namespace guit {

/// Smart pointer that owns a GObject.
/// @See GObject for details on automatic memory management.
template <typename T> class gptr {
public:
  gptr() noexcept : p_{} {}
  gptr(T* p) : p_(p)  {gaddPtr(p_,true);}
  gptr(T& p) : p_(&p) {gaddPtr(p_,true);}
  gptr(gptr const& b)  : p_(b.p_) {gaddPtr(p_,true);}
  gptr(gptr const&& b) : p_(b.p_) {gaddPtr(p_,true);}
  template <typename B> gptr(gptr<B> const& b)  : p_(b.p_) {gaddPtr(p_,true);}
  template <typename B> gptr(gptr<B> const&& b) : p_(b.p_) {gaddPtr(p_,true);}

  ~gptr() {gremovePtr(p_);}

  gptr& operator=(T* p) {reset(p);  return *this;}
  gptr& operator=(T& p) {reset(&p); return *this;}
  gptr& operator=(gptr<T> const& b)  {reset(b.p_); return *this;}
  gptr& operator=(gptr<T> const&& b) {reset(b.p_); return *this;}
  template <class B> gptr& operator=(B* p) {reset(p); return *this;}
  template <class B> gptr& operator=(gptr<B> const& b)  {reset(b.p_); return *this;}
  template <class B> gptr& operator=(gptr<B> const&& b) {reset(b.p_); return *this;}

  /// resets pointer.
  void reset() {gptr().swap(*this);}
  void reset(T* b) {gptr(b).swap(*this);}

  /// swaps pointees.
  void swap(gptr& b) noexcept {T* tmp = p_; p_ = b.p_; b.p_ = tmp;}

  /// returns number of references of pointee.
  unsigned int useCount() const noexcept {return p_ ? p_->refcount_ : 0;}

  /// returns pointee.
  T* get() const noexcept {return p_;}

  /// implicit type conversion: returns pointee.
  operator T*() const noexcept {return p_;}

  /// dereference: returns pointee.
  T* operator->() const noexcept {return p_;}

  /// dereference: returns reference to pointee.
  T& operator*() const noexcept {return *p_;}

  /// returns reference for Active Expressions (GExpr).
  auto operator~() const {return ~(*p_);}

  /// returns false if pointee is null.
  bool operator!() const noexcept {return p_ == nullptr;}

  /// increments then returns value of pointee.
  auto& operator++(int) const {return (*p_)++;}

  /// decrements then returns value of pointee.
  auto& operator--(int) const {return (*p_)--;}

  /// adds Gadget or Prop to pointee.
  auto& operator<<(const char* child) const {return *p_ << child;}
  auto& operator<<(GString const& child) const {return *p_ << child;}
  auto& operator<<(GStrings const& child) const {return *p_ << child;}
  auto& operator<<(Gadget& child) const {return *p_ << child;}
  auto& operator<<(Gadget* child) const {return *p_ << child;}
  auto& operator<<(GProp& child) const {return *p_ << child;}
  auto& operator<<(GProp* child) const {return *p_ << child;}
  auto& operator<<(class GPropList const& list) const {return *p_ << list;}

  /// binds pointee.
  template <typename C> auto& operator<<=(C& from) {return *p_ <<= from;}

  /// create state transition.
  template <typename C> auto& operator>>(C& to) {return *p_ >> to;}

protected:
  template <class X> friend class gptr;
  T* p_{};
};

template<class T>
inline bool operator==(gptr<T> const & a, std::nullptr_t b) {
  return a.get() == b;
}

template<class T>
inline bool operator!=(gptr<T> const & a, std::nullptr_t b) {
  return a.get() != b;
}

template<class T, class U>
inline bool operator==(gptr<T> const & a, U const * b) {
  return a.get() == b;
}

template<class T, class U>
inline bool operator!=(gptr<T> const & a, U const * b) {
  return a.get() != b;
}

template<class T, class U>
inline bool operator==(gptr<T> const & a, U const & b) {
  return a.get() == &b;
}

template<class T, class U>
inline bool operator!=(gptr<T> const & a, U const & b) {
  return a.get() != &b;
}

template<class T, class U>
inline bool operator==(gptr<T> const & a, gptr<U> const & b) {
  return a.get() == b.get();
}

template<class T, class U>
inline bool operator!=(gptr<T> const & a, gptr<U> const & b) {
  return a.get() != b.get();
}



template <typename T> class gnew : public gptr<T> {
public:
  gnew() : gptr<T>(new T) {}

  template <typename A>
  gnew(A arg) : gptr<T>(new T(arg)) {}
};



/// Smart pointer that protects a GObject from autodeletion whithout owning it.
/// gprotect<> prevents gptr<> pointers from immediately destroying an object:
/// - if all gptr<> pointers pointing to the object have been reset, the object is destroyed
///  when the gprotect<> is reset
/// - if no gptr<> points to the object or if there are still gptr<> pointers pointing to it,
///  nothing happens when the gprotect<> is reset
///
template <typename T> class gprotect : public gptr<T> {
  void init(T* p) {this->p_ = p; gaddPtr(p,false);}
public:
  gprotect() noexcept {}
  gprotect(T* p) {init(p);}
  gprotect(T& p) {init(&p);}
  gprotect(gprotect const& b) {init(b.p_);}
  gprotect(gprotect const&& b) {init(b.p_);}
  template <typename B> gprotect(gprotect<B> const& b) {init(b.p_);}
  template <typename B> gprotect(gprotect<B> const&& b) {init(b.p_);}

  gprotect& operator=(T* p) {reset(p);  return *this;}
  gprotect& operator=(T& p) {reset(&p); return *this;}
  gprotect& operator=(gprotect<T> const& b)  {reset(b.p_); return *this;}
  gprotect& operator=(gprotect<T> const&& b) {reset(b.p_); return *this;}
  template <class B> gprotect& operator=(B* p) {reset(p); return *this;}
  template <class B> gprotect& operator=(gprotect<B> const& b)  {reset(b.p_); return *this;}
  template <class B> gprotect& operator=(gprotect<B> const&& b) {reset(b.p_); return *this;}

  void reset() {gprotect().swap(*this);}
  void reset(T* b) {gprotect(b).swap(*this);}
  void swap(gprotect& b) noexcept {T* tmp = this->p_; this->p_ = b.p_; b.p_ = tmp;}
};

}
#endif
