//
//  guit Sockets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <thread>
#include <list>
#include <csignal>
#include <gapp.hpp>
#include <gnotifier.hpp>
#include <gbind.hpp>
#include <gstring.hpp>
#include <gsocket.hpp>
#include <gtrigger.hpp>
#include <gtext.hpp>
#include <gint.hpp>
#include <gnatsocket.hpp>
namespace guit {

namespace impl {

class GSocketBase {
public:
  virtual ~GSocketBase() = default;
  virtual int start(GString const& hostname, int port) = 0;
  virtual void sendData(GString const&) = 0;
  bool ready_{false};
};


class GSocketCnx : public GSocketBase {
  GSocket* mainsocket_;
  std::thread thread_;
  GNatSocket& sock_;
  GNatSocketBuffer& sockbuf_;
  bool receiving_{};
  int attempts_{};
  
public:
  GSocketCnx(GSocket* mainsocket, GNatSocket& socket) :
  mainsocket_(mainsocket),
  sock_(socket),
  sockbuf_(*new GNatSocketBuffer(sock_)) {
  }
  
  ~GSocketCnx() {
    sock_.close();
    if (thread_.joinable()) thread_.join();
    delete &sockbuf_;
    delete &sock_;
  }
  
  int start(GString const& hostname, int port) override {
    int stat = sock_.connect(hostname, port);
    if (stat >= 0) {  // OK
#if !defined(_WIN32)
      std::signal(SIGPIPE, SIG_IGN);  // ignore nasty SIGPIPEs
#endif
      startThread();
    }
    return stat;
  }
  
  void startThread() {
    ready_ = true;
    thread_ = std::thread([this]{receive();});
  }
  
  void sendData(GString const& data) override {
    // receiving_: do not send message if this message is being received
    // by the same socket and thus calls the send socket callback otherwise
    // we'll get an infinite loop
    if (!ready_  || (receiving_ && mainsocket_->dontsendback_)) return;
    if (sockbuf_.writeLine(data) < 0) {
      disconnected("could not send message: "+ data);
    }
  }
  
  void receive() {
    while (true) {
      GString message;
      
      if (sockbuf_.readLine(message) < 0) {
        disconnected("could not receive message");
        return;  // end of thread
      }

      if (!message.empty()) attempts_ = 0;
      else if (++attempts_ >= mainsocket_->maxattempts_) {
        disconnected("no data received");
        return;  // end of thread
      }
      
      auto& sockmap_ = mainsocket_->channels_;
      //cerr << "recv: '" <<message <<"'"<<endl;

      // send all data to channe * if it has been created
      auto it2 = sockmap_.find("*");
      if (it2 != sockmap_.end()) {
        auto* chan = it2->second.get();
        gexec([this,chan,message]{receiveCB(*chan, message);});
      }

      GString name, value;
      auto n = strsplit(message, name, value,  ":");
      if (n < 1) continue;

      // if name not found, nothing done because this data is probably not useful
      // for this client (en fait il fudrait n'envoyer que si le channel
      // est enregistree sur le client)
      auto it = sockmap_.find(name);
      if (it != sockmap_.end()) {
        auto* chan = it->second.get();
        // executed in main thread
        gexec([this,chan,value]{receiveCB(*chan, value);});
      }
    }
  }
  
  void receiveCB(GSocketChannel& chan, GString const& str) {
    receiving_ = true;
    if (chan.recv_) chan.recv_->set(str);
    if (!chan.props_.empty()) chan.props_.fire(&chan);
  }
  
  void disconnected(GString const& msg) {
    ready_ = false;
    mainsocket_->disconnected(this, msg);
    sock_.close();
  }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GSocketServer : public GSocketBase {
  GSocket* socketobj_;
  std::thread thread_;
  GNatServerSocket& servsock_;
  std::list<GSocketCnx*> cnxs_;
  int attempts_{};
  
public:
  GSocketServer(GSocket* socketobj) :
  socketobj_(socketobj),
  servsock_(*new GNatServerSocket) {
  }
  
  ~GSocketServer() {
    servsock_.close();
    if (thread_.joinable()) thread_.join();
    for (auto* it : cnxs_) delete it;
    delete &servsock_;
  }
  
  int start(GString const&, int port) override {
    int stat = servsock_.bind(port);
    if (stat >= 0) {   // OK
#if !defined(_WIN32)
      signal(SIGPIPE, SIG_IGN);  // ignore nasty SIGPIPEs
#endif
      ready_ = true;
      thread_ = std::thread([this]{accept();});
    }
    return stat;
  }
  
  void sendData(GString const& data) override {
    if (!ready_) return;
    //std::cout << "- SocketServer = " << data << endl;
    for (auto& it : cnxs_) it->sendData(data); // que connexions actives!
  }
  
  void accept() {
    while (true) {
      if (auto* socket = servsock_.accept()) {
        attempts_ = 0;
        //cerr << "input connection " << socket << endl;
        auto* c = new impl::GSocketCnx(socketobj_, *socket);
        cnxs_.push_back(c);
        
        // updates all channels by sending their current value for updating
        // the new client (en fait ne devrait etre envoyé qu'a ce client)
        c->ready_ = true;
        socketobj_->update();
        c->startThread();
      }
      else {
        //cerr << "input connection failed!" << endl;
        if (++attempts_ >= socketobj_->maxattempts_) {
          disconnected("input connection failed");
          return;  // end of thread
        }
      }
    }
  }
  
  void disconnected(GString const& msg) {
    ready_ = false;
    socketobj_->disconnected(this, msg);
    servsock_.close();
  }
};

} // endof impl

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GSocket::GSocket() :
status_(*new GInt(GSocket::Disconnected)),
impl_{}
{}

GSocket::~GSocket() {
  *status_ = GSocket::Disconnected;
  delete impl_;
  clear();
}

GInt& GSocket::status() {
  return *status_;
}

void GSocket::reset() {
  mode_ = Unstarted;
  *status_ = Disconnected;
  delete impl_;
  impl_ = nullptr;
}

bool GSocket::startAsServer(int port) {
  if (mode_ == Client) reset();
  if (status_->value() > 0) return true;   // already started
  
  if (!impl_) impl_ = new impl::GSocketServer(this);
  int stat = impl_->start("", port);
  
  if (stat < 0) {
    *(status_) = stat;
    reset();
    return false;
  }
  else {
    mode_ = Server;
    *(status_) = Connected;
    return true;
  }
}

bool GSocket::startAsClient(GString const& hostname, int port) {
  if (mode_ == Server) reset();
  if (status_->value() > 0) return true;  // already started
  
  if (!impl_) impl_ = new impl::GSocketCnx(this, *new GNatSocket);
  int stat = impl_->start(hostname, port);
  
  if (stat < 0) {
    *status_ = stat; // InvalidSocket = -3, UnknownHost = -2, CannotReach = -1
    reset();
    return false;
  }
  else {
    mode_ = Client;
    *(status_) = Connected;
    return true;
  }
}

void GSocket::disconnect() {
  reset();
  *status_ = GSocket::Disconnected;
}

void GSocket::sendData(GString const& data) {
  if (impl_ && status_->value() == Connected) {
    impl_->sendData(data);
  }
}

void GSocket::disconnected(impl::GSocketBase* impl, GString const& msg) {
  if (impl == impl_) {   // otherwise, connection with client, dont disconnect
    *status_ = Disconnected;
    //error("socket","disconnected: " + msg);
  }
}

// updates all channels by sending their current value, useful when connecting
// a client for updating it (en fait ca ne devrait etre envoyé qu'a ce client)
void GSocket::update() {
  if (impl_ && status_->value() == Connected) {
    for (auto& it : channels_) it.second->sendValue();
  }
}

// removes notifiers on all channels
void GSocket::clear() {
  for (auto& it : channels_) {
    // remove notifier that fires socket
    if (it.second->sender_) it.second->sender_->removeNotifiers(this);
  }
}

GSocketChannel& GSocket::operator[](GString const& name) {
  auto& chan = channels_[name];
  if (!chan) chan = new GSocketChannel();
    chan->name_ = name;        // A REVOIR
    chan->sock_ = this;
    return *chan;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

namespace impl {

// notifies the socket when prop value is changed
class GSocketNotifier : public GNotifier {
  GSocketChannel& chan_;
public:
  GSocketNotifier(GSocketChannel& c) : chan_(c) {}
  void* receiver() override {return chan_.socket();}
  void fire(GEvent&) override {chan_.sendValue();}
};
}

void GSocketChannel::sendValue() {
  if (sender_) sock_->sendData(name_ + ":" + sender_->stringValue());
}

GSocketChannel& GSocketChannel::operator=(GString const& data) {
  if (name_ == "*") sock_->sendData(data);
  else sock_->sendData(name_ + ":" + data);
  return *this;
}

GSocketChannel& GSocketChannel::operator<<=(GVarProp& sender) {
  (*sock_)[name_].sender_ = &sender;
  // adds model notifier that fires socket
  // IMPORTANT: the notifier must be added at the beginning of the prop list
  // because other callbacks migth modify the value of the prop => the notifier
  // must send the value before thus addBegin() rather than add()
  sender.addBegin(*new impl::GSocketNotifier(*this));
  return *this;
}

GSocketChannel& GSocketChannel::operator>>=(GVarProp& receiver) {
  (*sock_)[name_].recv_ = &receiver;
  return *this;
}

GSocketChannel& operator<<=(GVarProp& receiver, GSocketChannel& chan) {
  (*chan.sock_)[chan.name_].recv_ = &receiver;
  return chan;
}

GSocketChannel& GSocketChannel::operator^=(GVarProp& prop) {
  operator>>=(prop);
  operator<<=(prop);
  return *this;
}

GSocketChannel& GSocketChannel::operator<<(GFunction const& fun){
  //props_.addEnd(*new GCallbackNoArg(GTrigger::action, fun));
  props_.addEnd(on::action / fun);
  return *this;
}

}
