//
//  Strings
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_String_hpp
#define Guit_String_hpp
#include <gtypes.hpp>
namespace guit {

/// Removes beginning and trailing whitespaces (and quotes optionally).
/// If _remquotes_ is true, beginning and ending (simple or double) quotes are also removed
/// (but whitespaces inside quotes are kept).
void strtrim(GString& str, bool remquotes = false);


enum struct SplitMode {Raw = 0, Trim = 1, CheckQuotes = 2, RemQuotes = 3};

/** Splits a string into 2 tokens or a vector of strings.
 *  The first version searches for the first delimiter in _str_ and returns the substrings before (_left_)
 *  and after (_right_) this delimiter.
 *  The second version searches for the first _limit_ delimiters in _str_ and returns the resulting
 *  substrings as a vector (_tokens_). Searches for all delimiters If _limit_ is <= 0 (the default).
 *
 *  All characters in _delimiters_ are taken as delimiters. If _delimiters_ is empty (the default),
 *  whitespaces are taken as delimiters (using std::isspace()).
 *
 *  _mode_ can be:
 *  - SplitMode::Trim (the default): whitespaces around tokens are removed
 *  - SplitMode::Raw: whitespaces are not removed
 *  - SplitMode::CheckQuotes : delimiters are ignored inside (simple or double) quotes.
 *   Quotes are not removed, but whitespaces outside quotes are removed.
 *  - SplitMode::RemQuotes : same as CheckQuotes but quotes are removed. Whitespaces
 *   outside quotes are removed, but whitespaces inside quotes are kept.
 *
 *  Returns the number of tokens:
 *  - The second version returns _tokens_.size()
 *  - The first version returns 0 (empty _str_), 1 (no delimiter or second token is empty) or 2
 *   (second token non-empty). "empty" means after trimming when this option is set.
 *  - If not null, the optional _found_delim_ argument contains the delimiter, or 0 (no delimiter
 *   but _str_ non empty), or -1 (empty _str_)
 *  @code
 *      GString left, right;
 *      strsplit(" name : Bob ", left, right, ":");                      // [1]
 *      strsplit(" name : Bob ", left, right, ":", SplitMode::Raw);      // [2]
 *  @endcode
 *  - [1] left = "name" and right = "Bob"   (tokens are trimmed).
 *  - [2] left = " name " and right = " Bob  "  (tokens are not trimmed).
 *  @code
 *      GStrings tokens;
 *      strsplit("res one\ttwo", tokens);                                // [1]
 *      strsplit("res one\ttwo", tokens, " ");                           // [2]
 *      strsplit(" res = one + two ", tokens, "=+");                     // [3]
 *      strsplit(" res = one + two ", tokens, "=+", 0, SplitMode::Raw);  // [4]
 *  @endcode
 *  - [1] all whitespaces are delimiters: tokens = {"res", "one", "two"}
 *  - [2] only ' ' (space char) is a delimiter: tokens = {"res", "one\ttwo"}
 *  - [3] '=' and '+' are delimiters: tokens = {"res", "one", "two"}
 *  - [4] same as [3] except that the tokens are not trimmed: tokens = {"  res ", "  one  ", "  two " }
 *
 * @note GString is an alias for std::string and GStrings for std::vector<std::string>
 */
/// @{
size_t strsplit(GString const& str,
                GString& left, GString& right,
                GString const& delimiters = "",
                SplitMode mode = SplitMode::Trim);

size_t strsplit(GString const& str,
                GStrings& tokens,
                GString const& delimiters = "",
                int limit = 0,
                SplitMode mode = SplitMode::Trim);
/// @}

/** Splits a string into a vector of tokens according to a Regex.
 *  Same as strsplit() except that _str_ is split according to a regular expression:
 *  @code
 *      GStrings tokens;
 *      regsplit(" res = one + two ", tokens, "=|+"               // [1]
 *      regsplit(" res = one + two ", tokens, "=|+", 0, false);   // [2]
 *  @endcode
 *  - [1] tokens = {"res", "one", "two"}   (tokens are trimmed).
 *   Note the | in "=|+" because _regex_ is a Regex.
 *  - [2] tokens = {" res ", " one ", "  two "}    (tokens are not trimmed).
 *  @note This function is (much) slower than the other split() functions.
 */
size_t strregsplit(GString const& str,
                   GStrings& tokens,
                   GString const& regex,
                   int limit = 0,
                   bool trim = true);

/// Retreives the next token.
/// - _token_ will contain the next token before one of the _delimiters_ or the end of the string.
/// - returns the delimiter, or 0 (no delimiter but _str_ non empty), or -1 (empty _str_)
/// - see strsplit() for possible _mode_s.
int strtoken(GString const& str,
             GString& token,
             size_t& from,
             GString const& delimiters = "",
             SplitMode mode = SplitMode::Trim);

/// Lowercases the string (relies on toupper(int)).
GString strlower(GString const& str);

/// Uppercases the string (relies on tolower(int)).
GString strupper(GString const& str);

/// Returns true if the string is empty or contains only whitespaces.
bool strblank(GString const& str);

/// Compares two strings ignoring case (relies on tolower()).
/// @{
bool strequal(GString const& str1, GString const& str2, bool ignore_case);
int strcompare(GString const& str1, GString const& str2, bool ignore_case);
/// @}

/** Converts a C string into a numeric ID that can be used in switch blocks.
 * By default, the returned ID relies on the first 8 characters of the string.
 * If _last_ is provided, the ID relies on the last 8 characters before _last_.
 * @code
 *   int getRGB(const char* color) {
 *     switch (strid(color)) {
 *       case strid("red"):   return 0xf00;
 *       case strid("green"): return 0x0f0;
 *       case strid("blue"):  return 0x00f;
 *       default: return 0;
 *     }
 *   }
 * @endcode
 */
inline constexpr uint64_t strid(const char* str, int last = 8, uint64_t id = 0) {
  return (!*str || last==0) ? id : strid(str+1, last-1, (id<<8) + uint64_t(*str));
}

/// Converts a C++ string into a numeric ID that can be used in switch blocks.
/// @see strid(const char *)
inline uint64_t strid(GString const& str, int last = 8) {
  return strid(str.data(), last, 0);
}

/// converts position of character (UTF8 code point) to index (byte position) and vice-versa.
/// - each UTF8 character  can occupy 1 to 4 bytes.
/// - returns position of last byte/character if first argument is < 0.
/// @{
ssize_t strPosToIndex(GString const& str, ssize_t charpos, bool richtext);
ssize_t strIndexToPos(GString const& str, ssize_t index, bool richtext);
/// @}

/// returns index (byte position) of next or previous character.
/// - each UTF8 character  can occupy 1 to 4 bytes.
/// @{
size_t strNextIndex(GString const& str, ssize_t index, bool richtext);
size_t strPreviousIndex(GString const& str, ssize_t index, bool richtext);
/// @}

}
#endif
