//
// Position and z-layer of a Gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <gpos.hpp>
#include <gevent.hpp>
#include <gexpr.hpp>
#include <gbox.hpp>
#include <gpropspec.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP(Pos, GPropType::Pos)
GIMPLEMENT_PROP_TYPE(DragPos, GPos::Type())
GIMPLEMENT_PROP_TYPE(RelativePos, GPos::Type())

GPos& Pos(float x, float y, GPos::Layer layer) {return *new GPos(x,y,layer);}

GPosValue::GPosValue(GPoint const& pos, int l)
: GPoint(pos), layer(l) {
  if (layer < 0) layer = 0; else if (layer > 255) layer = 255;
}

GPosValue::GPosValue(float x, float y, int l)
: GPoint(x,y), layer(l) {
  if (layer < 0) layer = 0; else if (layer > 255) layer = 255;
}

GPosValue& GPosValue::operator=(GPoint const& p) {
  x = p.x; y = p.y;
  return *this;
}

bool GPosValue::operator==(GPosValue const& other) const {
  return x == other.x && y == other.y && layer == other.layer;
}

bool GPosValue::operator!=(GPosValue const& other) const {
  return x != other.x || y != other.y || layer != other.layer;
}

GPosValue GPosValue::blend(GPosValue const& p1, GPosValue const& p2, float coeff) {
  GPosValue res;
  float beta = 1.f - coeff;
  res.x = coeff * p1.x + beta * p2.x;
  res.y = coeff * p1.y + beta * p2.y;
  res.layer  = int(coeff * p1.layer  + beta * p2.layer);
  return res;
}

static GPropSpec posspec_{
  new GNumFieldSpec<GPosValue,float>(&GPosValue::x),
  new GNumFieldSpec<GPosValue,float>(&GPosValue::y),
  new GNumFieldSpec<GPosValue,int>(&GPosValue::layer)
};

bool gconvert(GString const& from, GPosValue& to) {
  to.x = to.y = 0.f;
  to.layer = 1;   // 1 needed for some reason
  int n = posspec_.toValue(from, &to);
  if (n <= 0) gerror("GPos","invalid keyword or value: " + from);
  return n > 0;
}

bool gconvert(GPosValue const& from, GString& to) {
  to = posspec_.toString(&from);
  return true;
}

// - - - - - - -

GPos::GPos() {
  value_.layer = 1;  // must be 1 !
}

GPos::GPos(GPoint const& pos, Layer layer)
: Super(GPosValue(pos, layer)) {}

GPos::GPos(float x, float y, Layer layer)
: Super(GPosValue(x, y, layer)) {}

GPos::GPos(GPos const& from)
: Super(from.value_) {}

GPos::GPos(GString const& value) {
  set(value);
}

GPos* GPos::clone(bool copy_value) const {
  return copy_value ? new GPos(*this) : new GPos;
}

void GPos::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Pos]) return;
  r.specif[GPropType::Pos] = specif;

  r.pos_ = value_;
}
  
void GPos::set(GPoint const& pos) {
  set(GPosValue(pos, value_.layer), false);
}

void GPos::set(float x, float y, Layer layer) {
  set(GPosValue(x, y, layer), false);
}

void GPos::set(float x, float y) {
  set(GPosValue(x, y, value_.layer),false);
}

void GPos::setX(float val) {
  set(GPosValue(val, value_.y, value_.layer),false);
}

void GPos::setY(float val) {
  set(GPosValue(value_.x, val, value_.layer),false);
}

void GPos::setLayer(int val) {
  set(GPosValue(value_.x, value_.y, val),false);
}

bool GPos::set(GProp const& p) {
  if (auto* c = p.toPos()) {
    set(c->value(),false);
    return true;
  }
  else if (p.isString()) {
    set(p.stringValue());
    return true;
  }
  else {
    error("set(GProp)","Incompatible argument");
    return false;
  }
}

GPos& GPos::operator<<=(GTextExpr& expr) {
  bindImpl(expr);
  this->set(expr.stringValue());
  return *this;
}

void GPos::blend(GPos const& a, GPos const& b, float coeff) {
  set(GPosValue::blend(a.value_, b.value_, coeff),false);
}

bool GPos::blend(const GProp& a, const GProp& b, float alpha) {
  auto* p1 = a.toPos();
  auto* p2 = b.toPos();
  if (p1 && p2) {
    set(GPosValue::blend(p1->value(), p2->value(), alpha),false);
    return true;
  }
  else return false;
}

void GPos::updateGadget(Gadget& g) {
  g.layout_.activePos = (value_.layer > 0);
  g.layout_.passivePos = (value_.layer == 0);
  g.layout_.layer = uint8_t(value_.layer);

  // otherwise the pos is passive, it is changed but dont moves the gadget
  if (value_.layer > 0) {
    if (auto* par = g.getParent()) {
      g.moveImpl(value_.x + par->wx(), value_.y + par->wy(), true);
    }
    else if (g.isTopLevel()) {
      g.moveImpl(value_.x, value_.y, true);
    }
    //else g.moveImpl(value_.x, value_.y, true);  NON!!!
    g.update();
  }
}

bool GPos::onAdd(Gadget* g) {
  if (!g) return true;
  g->layout_.activePos = (value_.layer > 0);
  g->layout_.passivePos = (value_.layer == 0);
  //g->updateLayout();
  return true;
}

void GPos::onRemove(Gadget* g) {
  if (!g) return;
  g->layout_.activePos = g->layout_.passivePos = 0;
  g->layout_.layer = 0;
  //if (isMutable()) endNotifyAndAnim(g);  fait par Gadget::remove()
  g->updateLayout();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GDragPos::GDragPos(GString const& str)
: GPos(str) {}

GDragPos::GDragPos(GDragPos const& from)
: GPos(from), fixx_(from.fixx_), fixy_(from.fixy_) {}

GDragPos* GDragPos::clone(bool copy_value) const {
  return copy_value ? new GDragPos(*this) : new GDragPos;
}

bool GDragPos::onAdd(Gadget* g) {
  if (!g) return true;
  g->layout_.reshapeFrame = g->layout_.activePos = (value_.layer > 0);
  g->layout_.passivePos = (value_.layer == 0);
  g->eventmask_ = g->eventmask_ | GEventType::MouseButton | GEventType::MouseDrag;
  return true;
}

void GDragPos::onRemove(Gadget* g) {
  if (!g) return;
  g->layout_.activePos = g->layout_.passivePos = 0;
  g->layout_.layer = 0;
  g->layout_.reshapeFrame = 0;
  g->updateLayout();
}

void GDragPos::fire(GEvent& e) {
  GMouseEvent* me = e.toMouseEvent();
  Gadget& g = e.from();
  GBox* parent = g.getParent();

  if ( (e.eventMask() & GEventType::MouseDrag) == 0
      || g.layout_.activeSize      // GInterSize will do the job in this case
      // || (channel_ >= 0 && channel_ != e.channel())
      || !parent || !me
      )
    return;

  GPoint p;
  if (!fixx_) p.x = value_.x + me->dx();
  if (!fixy_) p.y = value_.y + me->dy();
  g.keepInside(*parent, p);
  if (p != value_) g.move(p);
  
  /*
  if ((e.eventMask() & GEventType::MouseDrag) != 0) {
    auto loc = GHandle::Inside;
    e.gadget().reshapeDrag(me->dx(), me->dy(), loc);
  }*/
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// NOTE: related_ est un gptr => continue d'exister tant que GRelativePos
// existe, on pourrait faire mieux!

GRelativePos::GRelativePos(Gadget& related, float xoffset, float yoffset, Layer layer)
: GRelativePos(related, GPoint(xoffset,yoffset), layer) {}

GRelativePos::GRelativePos(Gadget& related, GPoint const& pos, Layer layer)
: GPos(GPoint(), layer), related_(&related) {
  relatedpos_ = pos;
  cond_ = on::move / [this]{updatePos();};
  related << *cond_;
  updatePos();
}

GRelativePos::GRelativePos(GRelativePos const& from)
: GPos(from), related_(from.related_), cond_(from.cond_), relatedpos_(from.relatedpos_) {}

GRelativePos::GRelativePos(GString const& str) : GPos(str) {}

GRelativePos* GRelativePos::clone(bool copy_value) const {
  return copy_value ? new GRelativePos(*this) : new GRelativePos;
}

GRelativePos::~GRelativePos() {
  if (cond_) related_->remove(cond_);
}

void GRelativePos::updatePos() {
  set(related_->pos() + relatedpos_);
}

}
