//
// Types of properties (Props in short)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_PropType_hpp
#define Guit_PropType_hpp
#include <gobject.hpp>
namespace guit {


/// Base of Prop Types (kind of metaclass).
class GPropType {
public:
  // NOTE: props with a negative role are never equivalent (see GProp::isEquivalent()).
  enum Role {
    Modes = -8, States = -7, Holder = -6, Expr = -5, ExprWrapper = -4,
    Notifier = -3, Cond = -2, Abstract = -1,
    Align, Background, Border, Color, Flex, Flow, Font, Format, Gap, Hotkey,
    Icon, Margin, Pos, Size, Text, Tip, RenderCount,
    Bool, Number, Choice, Anim,
    Auto  // number is auto set
  };

  /// Return the name and the role of the proptype (depends of subclass).
  /// @{
  GString const& name() const {return name_;}
  int role() const {return role_;}
  /// @}

  /// Instantiates a Prop of this type (redefined in subclasses).
  /// @warning: Throws if the prop is not instantiable (classes deriving from GVarProp are instantiable).
  /// @{
  virtual GProp& newInstance() noexcept(false);
  virtual GProp& newInstance(GString const& arg) noexcept(false);
  /// @}

  /// Instantiates a Prop according to the specified type name.
  static GProp* createInstance(GString const& type_name, GString const& arg,
                               bool css_style = false);

  /// Returns the type corresponding to this name if it exists, null otherwise.
  static GPropType* findType(GString const& type_name);

  bool operator==(GPropType const& other) const;
  bool operator!=(GPropType const& other) const;

  /// The delete operator does nothing and should not be used.
  void operator delete(void*);

  /// @internal
  /// @{
  virtual ~GPropType() = default;
  GPropType(GString const& name, int role);
  GPropType(GString const& name, GPropType& parent);
  /// @}

protected:
  friend class GProp;
  friend class GPropList;

  GPropType(GPropType&) = delete;
  GPropType(GPropType&&) = delete;
  GPropType& operator=(GPropType const&) = delete;
  void addType();

  GString name_;
  mutable int role_{};
  GPropType* parent_{};
};


/// PropType of a Prop of class T.
template <class P, class Super>
class GPropType_ : public Super::GType {
public:
  using Value = P;

  P& newInstance() override {return *new P();}
  P& newInstance(GString const& arg) override {return *new P(arg);}

  /// Binds a prop to another prop.
  template <typename O>
  P& operator<<=(O& other) {P& p{(*this)()}; p <<= other; return p;}
    
  /// @internal.
  /// @{
  GPropType_(GString const& name, int role) : Super::GType(name, role) {}
  GPropType_(GString const& name, GPropType& parent) : Super::GType(name, parent) {}
  /// @}
};

}
#endif
