//
// Size of a Gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <gevent.hpp>
#include <gpropspec.hpp>
#include <gsize.hpp>
#include <gexpr.hpp>
#include <gwindow.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP(Size, GPropType::Size)
GIMPLEMENT_PROP_TYPE(DragSize, GSize::Type())

bool GDim::operator==(const GDim& other) const {
  return w == other.w && h == other.h;
}

bool GDim::operator!=(const GDim& other) const {
  return w != other.w || h != other.h;
}

std::ostream& operator<<(std::ostream& out, const GDim& d) {
  out << d.w << " "<< d.h;
  return out;
}

std::istream& operator>>(std::istream& in, GDim& d) {
  in >> d.w >> d.h;
  return in;
}

GSizeValue::GSizeValue()
//: GDim(-1, -1), wunit(GSize::Auto), hunit(GSize::Auto) {}
: wunit(GSize::Undef), hunit(GSize::Undef) {}

GSizeValue::GSizeValue(float w, float h, char wtype, char htype) :
GDim(w, h),
// NEEDED by GDragSize or some reason !!!!
wunit(w < 0 ? GSize::Auto : wtype),
hunit(h < 0 ? GSize::Auto : htype)
{}

GSizeValue::GSizeValue(GDim const& dim, char wtype, char htype) :
GDim(dim),
// NEEDED by GDragSize or some reason !!!!
wunit(w < 0 ? GSize::Auto : wtype),
hunit(h < 0 ? GSize::Auto : htype)
{}

bool GSizeValue::operator==(const GSizeValue& other) const {
  return
  w == other.w && h == other.h && wunit == other.wunit && hunit == other.hunit;
}

bool GSizeValue::operator!=(const GSizeValue& other) const {
  return
  w != other.w || h != other.h || wunit != other.wunit || hunit != other.hunit;
}

GSizeValue GSizeValue::blend(GSizeValue const& p1, GSizeValue const& p2, float coeff) {
  GSizeValue res;
  float beta = 1.f - coeff;
  res.w = coeff * p1.w + beta * p2.w;
  res.h = coeff * p1.h + beta * p2.h;
  return res;
}

// - - - - - -

static GFieldSpecValues sizevalues_ {
  {"a",  GSize::Auto, true},
  {"auto", GSize::Auto},
  {"%f", GSize::Pixels},
  {"%px",GSize::Pixels},
  {"%c", GSize::Chars},
  {"%%", GSize::Percent},
  {"*",  GSize::Undef}
};

static GPropSpec sizespec_{
  new GListFieldSpec2<GSizeValue,GSizeValue,char,float>
  (&GSizeValue::wunit, &GSizeValue::GDim::w, sizevalues_),

  new GListFieldSpec2<GSizeValue,GSizeValue,char,float>
  (&GSizeValue::hunit, &GSizeValue::GDim::h, sizevalues_)
};


bool gconvert(GString const& from, GSizeValue& to) {
  to.w = to.h = 0;
  to.wunit = to.hunit = GSize::Undef;

  int n = sizespec_.toValue(from, &to);
  // cerr <<"convert: '"<< from <<"' to: " << to.w <<" "<< to.h
  // <<" / " << int(to.wunit)  <<"  " << int(to.hunit) << endl<< endl;

  if (to.wunit == GSize::Pixels) {
    if (to.w < 0) {to.w = 0; to.wunit = GSize::Auto;}   // needed!
  }
  if (to.hunit == GSize::Pixels) {
    if (to.h < 0) {to.h = 0; to.hunit = GSize::Auto;}   // needed!
  }
  //else if (to.wunit == GSize::Percent) {
  //  if (to.w < 0.f) to.w = 0.f; else if (to.w > 100.f) to.w = 100.f;
  //}
  //else if (to.hunit == GSize::Percent) {
  //  if (to.h < 0.f) to.h = 0.f; else if (to.h > 100.f) to.h = 100.f;
  //}
  return n > 0;
}

bool gconvert(GSizeValue const& from, GString& to) {
  // cerr <<"convert: " << from.w <<" "<< from.h
  // <<" / " << int(from.wunit)  <<" " << int(from.hunit) <<endl;
  to = sizespec_.toString(&from);
  // cerr<<"' to: " << to
  // << endl<< endl;
  return true;
}

// - - - - -

GSize::GSize() = default;

GSize::GSize(float w, float h, Unit wtype, Unit htype)
: Super(GSizeValue(w, h, wtype, htype)) {}

GSize::GSize(GDim const& dim, Unit wtype, Unit htype)
: Super(GSizeValue(dim, wtype, htype)) {}

GSize::GSize(GSize const& from) : Super(from.value_) {}

GSize::GSize(GString const& format) {set(format);}

GSize* GSize::clone(bool copy_value) const {
  return copy_value ? new GSize(*this) : new GSize;
}

void GSize::set(float w, float h, Unit wtype, Unit htype) {
  set(GSizeValue(w, h, wtype, htype), false);
}

bool GSize::set(GProp const& p) {
  if (auto* c = p.toSize()) {
    set(c->value(), false);
    return true;
  }
  else if (p.isString()) {
    set(p.stringValue());
    return true;
  }
  else {
    error("set(GProp)","incompatible argument");
    return false;
  }
}

void GSize::blend(GSize const& a, GSize const& b, float coeff) {
  set(GSizeValue::blend(a.value_, b.value_, coeff) ,false);
}

bool GSize::blend(const GProp& a, const GProp& b, float alpha) {
  auto* p1 = a.toSize();
  auto* p2 = b.toSize();
  if (p1 && p2) {
    set(GSizeValue::blend(p1->value(), p2->value(), alpha), false);
    return true;
  }
  else return false;
}

void GSize::apply(Gadget* g, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Size]) return;
  r.specif[GPropType::Size] = specif;
  
  if (value_.wunit != Undef && value_.hunit != Undef) {
    r.size_ = value_;
  }
  else if (value_.wunit != Undef) {
    r.size_.w = value_.w;
    r.size_.wunit = value_.wunit;
  }
  if (value_.hunit != Undef) {
    r.size_.h = value_.h;
    r.size_.hunit = value_.hunit;
  }
}

void GSize::updateGadget(Gadget& g) {
  //cerr << "updategadget " << this << " = " <<value_.w << " " << value_.h<<endl;

  // g.updateLayout entraine une boucle oo d'evenements Resize et de requetes resize()
  // g.updateLayout();

  // !!! TODO: si la taille n'est pas en pixel !!!
  // ce code remmet à jour g.frame_ qunad size est change, il est indispensable
  // sinon size.set() n'a pas d'effet mais il ne marche que pour size en pixels
  if (value_.wunit == GSize::Pixels && value_.w >= 0) {
    g.frame_.w = value_.w;
  }
  if (value_.hunit == GSize::Pixels && value_.h >= 0) {
    g.frame_.h = value_.h;
  }
  
  if (auto* win = g.toWindow()) {
    win->natwin_->resize(g.frame_.w, g.frame_.h);
    win->pack(nullptr);
  }
  if ((g.eventmask_ & GEventType::Resize) != 0) {
    GEvent e(GTrigger::resize, GCore::elapsedTime());
    g.doFire(e);
  }
  g.update();
}

bool GSize::onAdd(Gadget* g) {
  if (!g) return true;
  g->layout_.activeSize = 1;
  g->updateLayout();
  return true;
}

void GSize::onRemove(Gadget* g) {
  if (!g) return;
  g->layout_.activeSize = 0;
  // remove events
  g->updateLayout();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GDragSize::GDragSize(GString const& str) : GSize(str) {}

GDragSize::GDragSize(GDragSize const& from) : GSize(from),
fixw_(from.fixw_), fixh_(from.fixh_), place_(from.place_) {}

GDragSize* GDragSize::clone(bool copy_value) const {
  return copy_value ? new GDragSize(*this) : new GDragSize();
}

bool GDragSize::onAdd(Gadget* g) {
  if (!g) return true;
  g->layout_.reshapeFrame = g->layout_.activeSize = 1;
  g->eventmask_ = g->eventmask_ | GEventType::MouseButton | GEventType::MouseDrag;
  return true;
}

void GDragSize::onRemove(Gadget* g) {
  if (!g) return;
  g->layout_.activeSize = 0;
  g->layout_.reshapeFrame = 0;
  // ... remove events
  g->updateLayout();
}

void GDragSize::fire(GEvent& e) {
  // cerr << "GDragSize::fire " <<  this << endl;
  GMouseEvent* me = e.toMouseEvent();

  if (//NO (e.eventMask() & GEventType::MouseDrag) == 0
      // || (channel_ >= 0 && channel_ != e.channel())
      !me)
    return;

  // TODO !!!!
  // fixw_ fixh_ pas traités !!

  if ((e.eventMask() & GEventType::MouseDrag) != 0) {
    // cerr << "drag "<<  me->dx() << " "<< me->dy()
    // << " size " <<  e.gadget().dim() << endl;
    e.from().handleDrag(me->dx(), me->dy(), place_);
  }
  else if ((e.eventMask() & GEventType::MouseButton) != 0) {
    if (e.trigger() == on::mousedown) {
      place_ = e.from().handleDown(me->wpos(), me->reshapeinter_);
    }
    else if (e.trigger() == on::mouseup) {
      e.from().handleUp(me->wpos(), place_);
      place_ = {};
    }
  }
}

}
