//
// Size of a Gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitSize_hpp
#define GuitSize_hpp
#include <gvarprop.hpp>
namespace guit {

class GSizeValue : public GDim {
public:
  GSizeValue();
  GSizeValue(float w, float h, char wunit, char hunit);
  GSizeValue(GDim const&, char wunit, char hunit);

  static GSizeValue blend(GSizeValue const& a, GSizeValue const& b, float coeff);

  bool operator==(GSizeValue const&) const;
  bool operator!=(GSizeValue const&) const;

  char wunit{}, hunit{};
};

inline bool gconvert(GSizeValue const& from, GSizeValue& to) {to = from; return true;}
bool gconvert(GString const& from, GSizeValue& to);
bool gconvert(GSizeValue const& from, GString& to);


/** Controls the actual or preferred size of a Gadget.
 * width and height are the preferred size of the gadget. When specified as a string, the format is:
 *     format = "width heigh"
 * with each value being either:
 * - a or auto        : the size is auto-computed according to the gadget content
 * - 12.5 or 12.5p : the size in pixels (the value must be a positive float)
 * - 10c                : the size of as many 'G' characters (using gadget's font)
 * - 20%               : percent of the size of the gadget's parent (must be between 1% and 100%)
 * - *                     : undefined value (won't be taken into account)
 * @code
 *    auto& b1 = Button("Do It") << Size("20 auto");  // pixels, auto-computed
 *    auto& b2 = Button("Do It") << Size("20c a");    // characters, auto-computed
 * @endcode
 * @note The actual  size also depends on the GFlex prop of the gadget and on the GAlign prop of its parent.
*/
class GSize : public GVarProp_<GSizeValue> {
public:
  using Super = GVarProp_<GSizeValue>;
  enum Unit : int8_t {Undef, Pixels, Chars, Auto, Percent};

  GSize* toSize() override {return this;}
  GSize const* toSize() const override {return this;}

  GSize();
  GSize(float width, float height, Unit wunit = Pixels, Unit hunit = Pixels);
  GSize(GDim const&, Unit wunit = Pixels, Unit hunit = Pixels);
  GSize(GSize const&);
  GSize(GString const& format);
  GSize* clone(bool copy_value) const override;

  GSize& operator=(GSize const& from) {set(from); return *this;}
  GSize& operator=(GString const& format) {set(format); return *this;}

  using Super::set;
  void set(float width, float height, Unit wunit = Pixels, Unit hunit = Pixels);
  bool set(GProp const&) override;

  virtual void blend(GSize const& a, GSize const& b, float alpha);
  bool blend(GProp const& a, GProp const& b, float alpha) override;

  float w() const {return value_.w;}   ///< >= 0: preferred size or %, -1: auto
  float h() const {return value_.h;}

  /// Field of a GSize prop.
  using Field = GNumPropField<GSize,float>;

  /// Returns a field that can be bound.
  /// This field can be bound to another prop or prop field by gbind() or operator <<=.
  /// @{
  Field wField() {return Field(*this, value_.w);}
  Field hField() {return Field(*this, value_.h);}
  /// @}

  using GType = GPropType_<GSize,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  friend class GRender;
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void apply(Gadget*, GRender&, Specif) final;
  void updateGadget(Gadget&) override;
};

/// Returns a GSize created in the heap.
/// @{
GSize& Size();
GSize& Size(GString const& value);
/// @}



/// Draggable size: Makes  the Gadget interactively resizable by dragging its borders.
class GDragSize : public GSize {
public:
  GDragSize(float w, float h) : GSize(w, h, Pixels, Pixels) {}
  GDragSize(GDim const& dim) : GSize(dim, Pixels, Pixels) {}
  GDragSize(GDragSize const&);
  GDragSize(GString const& = "");
  GDragSize* clone(bool copy_value) const override;

  /*
   void fixW(bool val = true) {fixw_ = val;}
   void fixH(bool val = true) {fixh_ = val;}
   bool isWFixed() const {return fixw_;}
   bool isHFixed() const {return fixh_;}
   */

  using GType = GPropType_<GDragSize,GSize>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal.
  /// @{
  bool fixw_{}, fixh_{};
  GHandle place_{};
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void fire(GEvent&) override;
  /// @}
};

}
#endif
