//
// Base class of variable Props (eg GInt, FText, GColor, GFont)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_VarProp_hpp
#define Guit_VarProp_hpp
#include <gproplist.hpp>
namespace guit {

/// Base class of variable Props.
/// The value of these props can be changed:
/// - When their value changes, parent gadgets are automatically updated.
/// - Callbacks can be added to them (see operator<<())
/// - They can be bound together:
/// @code
///   auto& txt = Text();
///   auto& num = Float();
///
///   txt <<= num;   // (1) num updates text
///   num <<= txt;   // (2) txt updates num
///   txt ^= num;    // (3) txt and num are synced (order matters: see below!)
/// @endcode
/// Props can be bound both ways (1+2) or synced (3). Syncing is slightly different:
/// - not only the 2 props are bound both ways,
/// - but the right-hand prop constraints the left-hand prop.
///
/// For instance, (3) means that txt must verify the isValid() condition of num.
/// So, if num is constrained by a range (see GFloat::setRange()), txt will always
/// be in this range. Moreover, if txt is added to a TextField or a TextArea,
/// the user will be forced to enter a numerical value lying in this range:
/// @code
///   auto& txt = Text();
///   auto& num = Float();
///   num.setRange(-100, 100);
///   txt ^= num;
///   auto& tf = (TextField << txt);
/// @endcode
///
/// Finally, one can also just write:
/// @code
///   auto& num = Float();
///   num.setRange(-100, 100);
///   auto& tf = (TextField << num);
/// @endcode
/// An internal Text prop is then created and synced with num, so that this
/// code is equivalent to the previous except.
///
/// @see operator<<=() and operator^=().
///
class GVarProp : public GProp {
public:
  using IsVarProp = bool;
  using IsGadgetOrVarProp = bool;

  using BindTag = int8_t;

  ~GVarProp();

  GVarProp* toVarProp() override {return this;}

  GVarProp* toMutableProp() override {
    return omodes_.immutable ? nullptr : this;
  }

  bool isMutable() const override {return !omodes_.immutable;}
  void setImmutable() override {omodes_.immutable = 1;}

  /// Changes value.
  virtual void set(GString const& value) = 0;

  /// Changes value.
  /// Does nothing if the type of _from_ is not convertible to the type of this prop.
  /// @see set() and operator=() methods in subclasses
  virtual bool set(GProp const& value) = 0;

  /// Changes value by blending the value of two props.
  /// Does nothing if the type of _from1_ or _from2_ is not convertible to the type of *this* prop.
  virtual bool blend(GProp const& value1, GProp const& value2, float mix);

  /// Reads the value of the prop from a file or a stream.
  /// @{
  bool read(GString const& filename);
  virtual void read(std::istream& in);
  friend std::istream& operator>>(std::istream& in, GVarProp&);
  /// @}

  /// Unbinds this prop.
  /// - the first version unbinds another prop (note that it doesn't unbind Prop fields).
  /// - the second version unbinds bindings according to its _bindtag_ argument:
  ///  # if bindtag is >= 0, unbinds bindings performed with the same bindtag (default by 0, see bind())
  ///  # if bindtag is < 0, unbinds all bindings with a value >= bindtag
  /// Importantly, these methods unbinds props both ways if the binding was done both ways
  /// (e.g. by using sync())
  /// @{
  void unbind(GVarProp& sender);
  void unbind(GVarProp::BindTag bindtag = 0);
  /// @}

  /// Blinks gadget parents of the prop.
  /// Shows a warning if msg is not empty.
  void blink(GString const& msg = "", GTime duration = 500);

  /// updates gadget parents of the prop.
  void update();

  /// performs this function on gadget parents.
  void perform(std::function<void(Gadget&)> fun);

  /// Returns callbacks and bound props.
  GPropList const& props() const {return props_;}

  virtual void addFun(GFunction const& fun);
  virtual void add(GProp&);
  virtual void add(GCond&);
  virtual void addBegin(GProp&);
  virtual void remove(GProp&);
  virtual void removeNotifiers(void* object) override;
  virtual void removeAll();   ///< Removes callbacks and notifiers.
  virtual void removeFromGadgets();

  /// @internal
  /// @{
  bool changing() const {return pmodes_.changing;}
  void changing(bool val) {pmodes_.changing = val;}
  void bindImpl(GExpr& sender);
  void bindImpl(GVarProp& sender, GFunction fun, BindTag);
  void bindImpl(GVarProp& sender, class GChangeNotifier&, BindTag, bool recv_is_master);
  void fireCallbacks(GEvent&);
  void fireDefaultCallbacks();
  bool hasActiveNotifiers();
  void startNotify(Gadget&, std::function<void(Gadget&)> const& fun);
  void startNotifyGadget(Gadget&) override;
  void stopNotify(Gadget&);
  bool testMasterValue();
  void propagateChange();
  void immutableError(const char*);
  class GChangeNotifier* getMaster();
  struct {      // nb: ordre optimise pour gagner de la place memoire
    uint8_t changing:1, removing:1, isSet:1, hasMaster:1, autoReset:1;
  } pmodes_{};
  GPropList props_;
  /// @}
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Superclass of a Mutable Prop class (template for implementing them).
template <class T> class GVarProp_ : public GVarProp {
public:
  using ValueType = T;

  /// Creates a mutable prop.
  /// @{
  GVarProp_() = default;
  explicit GVarProp_(T const& value) {doChange(value);}
  /// @}

  /// Returns the value.
  /// @{
  T const& value() const {return value_;}
  T const& operator()() const {return value_;}
  /// @}

  /// Implicitely converts to value.
  //operator T const&() const {return value_;}

  /// Returns the value as a string.
  GString stringValue() const override {return guit::toString(value_);}

  /// Changes the value.
  /// Callbacks are fired and parent gadgets updated if:
  /// - the value is different from the value that was previously set (using set() or operator=)
  /// - _callalways_ is true, in which case they are always fired
  /// Does nothing and produces an warning  if the prop is immutable (see setImmutable())
  virtual void set(T const& value, bool callalways) {
    if (omodes_.immutable) immutableError("set");
    else if (doChange(value) || callalways)
      propagateChange();
  }

  /// Changes the value from a string.
  void set(GString const& str) override {
    T val{}; gconvert(str, val); set(val, false);
  }
  
  bool set(GProp const& value) override {
    if (auto* p = dynamic_cast<GVarProp_<T> const*>(&value)) {
      set(p->value(), false); return true;
    }
    else if (value.isString()) {
      set(value.stringValue()); return true;
    }
    else {
      error("set(GProp)","incompatible argument"); return false;
    }
  }

  bool operator==(T const& value) const {return value_ == value;}

  template <class V>
  bool operator==(GVarProp_<V> const& other) const {return value_ == other.value_;}


  virtual bool isValid(T const&) const {return true;}

  virtual bool doChange(T const& v) {
    if (v == value_ && pmodes_.isSet) return false;
    pmodes_.isSet = 1;
    value_ = v;
    return true;
  }

  /// @internal
  T value_{};
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Prop reference: allows referencing a GVarProp in an Active Expression.
/// Such a reference can be obtained obtained through the ~ operator of the prop.
/// The difference between GPropRef and GNumPropRef is that the later refers to a numeric variable.
/// @see GExpr.
/// @{
class GPropRef {
public:
  explicit GPropRef(GProp& prop) : prop_(prop) {}
  GProp& prop_;
};
class GNumPropRef {
public:
  explicit GNumPropRef(GProp& prop) : prop_(prop) {}
  GProp& prop_;
};
/// @}


template <class P> class GPropRef_ : public GPropRef {
public:
  explicit GPropRef_(P& prop) : GPropRef(prop) {}
};

template <class P> class GNumPropRef_ : public GNumPropRef {
public:
  explicit GNumPropRef_(P& prop) : GNumPropRef(prop) {}
};


/// Prop field reference: allows referencing a field of a GVarProp in an Active or Binding Expression.
/// Same function as GPropRef but for a GVarProp field. Can also be used to bind a Prop with a field
/// of another prop and vice versa (see gbind()).
/// Such a reference can be obtained obtained through the xxxField() methods of aProp (eg GColor::redField()).
/// The difference between GPropField and GNumPropField is that the later refers to a numeric field.
/// @see GExpr and gbind()
/// @{
template <class P, class F>
class GPropField {
public:
  GPropField(P& prop, F& field) : prop_(prop), field_(field) {}
  GPropRef operator~();
  P& prop_;
  F& field_;
};
template <class P, class F>
class GNumPropField : public GPropField<P,F> {
public:
  GNumPropField(P& prop, F& field) : GPropField<P,F>(prop,field) {}
  GNumPropRef operator~();
};
/// @}

}
#endif
