//
//  Guit Script Editor
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <gapp.hpp>
#include <gscript.hpp>
#include <gfile.hpp>
//#include <gbuilder.hpp>
#include <ginspector.hpp>
#include "scriptedit.hpp"
using namespace std;
using namespace guit;

ScriptEditor::ScriptEditor() :
GWindow("#script-editor Guit Script Editor"),
inspect(GInspector::instance())
{
  inspect.setInspectingCB( [this](int b) {inspectingCB(b);} );
  inspect.setInspectedCB( [this](Gadget* g) {inspectedCB(g);} );

  selection.addBegin(*new GBackground("orange"));

  // 'changed' is set to true when 'code' is changed
  code << [this]{changed = true;};

  // redirect errors to show them on 'errors'
  GApp::onError([this](GString const& ctx, GString const& msg, GObject const*){
    errors += "\n<c=red>" + ctx + ":</c> "+ msg;
  });

  codearea.setAttributes(&tokens);

  auto& opendial = OpenDialog("Open Script", ":").alertIf(changed);
  opendial << [&]{readFile(opendial.pathname());};

  auto& savedial = SaveDialog("Save Script", "~");
  savedial << [&]{writeFile(savedial.pathname());};

  auto& quitdial = QuitDialog("Qui","Changes will be lost").openOnlyIf(changed);

  // contains the source code
  auto& codebox = VBox("#codebox")
  << (HBox(".titlebox")
      << Label("<b>Source Code")
      << Filler
      << (Button("Example") << [&]{readFile(Example); changed = false;})
      << (Button("Build") << [this]{build();})
      << (Button("Inspector") << [this]{showInspector();})

      )
  << (ScrollBox
      << (codearea
          << code
          << Font("14px courier")
          << on::mousedown / [&](GEvent& e){inspectCB(e.asMouseEvent());}
          )
      );

  // contains errors if any
  auto& errorbox = VBox("#errorbox")
  << (HBox(".titlebox")
      << "<b>Errors"
      << Filler
      << (Button("Clear") << [this]{errors = "";} )
      )
  << (ScrollBox
      << (TextArea << errors).setRichText(true)
      );

  // main window
  *this
  << quitdial
  << (ToolBar
      << ("Open" << opendial)
      << ("Save" << savedial)
      )
  << codebox
  << errorbox;

  // disables inspection
  this->setInternal();
  opendial.setInternal();
  savedial.setInternal();
  quitdial.setInternal();
}


int main(int argc, const char **argv) {
  GApp app(argc, argv);
  app.inspector();
  app <<
  "#script-editor {icon: icon.png}"
  "#script-editor .titlebox {size: a a; flex: 1 0; margin: 0 0}"
  "#script-editor #codebox {border:squared; size: 65C 35C; flex: 1 1}"
  "#script-editor #codearea {font: 12px Courier}"
  "#script-editor #errorbox {border:squared; size: 55C 8C; flex: 1 1}"
  "#script-editor #controls {align: s c; size: a a; flex: 1 0; margin: 10 10 5 10}";

  return app.start(new ScriptEditor);
}


void ScriptEditor::showInspector() {
  inspect.show(true);
  inspect.raise();
}

/*
 void ScriptEditor::showBuider() {
 builder = GApp::getBuilder(true);
  builder->show();
  builder->raise();
}*/

void ScriptEditor::readFile(GString const& filename) {
  changed = false;
  reset();

  if (!(code.read(filename))) {
    code = "File not found: "+ filename;
  }
  else {
    *this << Title("Script File: " + basename(filename));
    changed = false;
    build();
  }
}

void ScriptEditor::writeFile(GString const& filename) {
  if (code.write(filename)) changed = false;
}

void ScriptEditor::build() {
  tokens.clear();
  errors = "";

  if (Gadget* g = GLoader::loadBuffer(code(), nullptr, &tokens)) {
    GWindow* win =  g->toWindow();
    // if  not a window create one
    if (!win) win = & (*new GWindow("script") << g);

    *win << new GPos(0, 800);
    win->dieOnClose();    // destroy window on close
    win->show();
  }
}

void ScriptEditor::reset() {
  selected_gadget = nullptr;
  clickedattr = nullptr;
  clickedprops.list().clear();
  tokens.clear();
}

void ScriptEditor::inspectingCB(int b) {
  if (b == 0) {
    if (clickedattr) clickedattr->props_ = clickedprops;
      selected_gadget = nullptr;
      clickedattr = nullptr;
      clickedprops.list().clear();
    //reset();
  }
  //update();
}

void ScriptEditor::inspectCB(GMouseEvent& e) {
  ssize_t caret = codearea.findCaretIndex(e.wpos());
  if (caret < 0) return;

  if (GTextAttr* attr = tokens.find(caret)) {
    if (attr->data_ && inspect.inspecting()) {
      if (Gadget* g = attr->data_->toGadget()) inspect.inspect(g);
    }
  }
}

void ScriptEditor::inspectedCB(Gadget* g) {
  if (!g) {
    //cerr << "\ngadgetSelectedCB: null g" << endl;
  }
  else {
   // cerr << "\ngadgetSelectedCB " << g->className() <<" "<< g << endl;
    if (g != selected_gadget) {
      selected_gadget = g;

      if (GTextAttr* attr = tokens.find(g)) {
        //cerr << "-- gadgetSelectedCB " << attr << endl;
        if (clickedattr) clickedattr->props_ = clickedprops;
        clickedattr = attr;
        clickedprops = attr->props_;
        clickedattr->props_ = selection;
        update();
      }
    }
  }
}

